# Copyright 2024 VAST Data Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import itertools
import unittest
from unittest import mock

import ddt
import netaddr

import manila.context as manila_context
import manila.exception as exception
from manila.share import configuration
from manila.share.drivers.vastdata import driver
from manila.share.drivers.vastdata import driver_util
from manila.tests import fake_share
from manila.tests.share.drivers.vastdata.test_rest import fake_metrics


@mock.patch(
    "manila.share.drivers.vastdata.rest.Session.refresh_auth_token",
    mock.MagicMock()
)
@ddt.ddt
class VASTShareDriverTestCase(unittest.TestCase):
    def _create_mocked_rest_api(self):
        # Create a mock RestApi instance
        mock_rest_api = mock.MagicMock()

        # Create mock sub resources with their methods
        subresources = [
            "views",
            "view_policies",
            "capacity_metrics",
            "quotas",
            "vip_pools",
            "snapshots",
            "folders",
        ]
        methods = [
            "list", "create", "update",
            "delete", "one", "ensure", "vips"
        ]

        for subresource in subresources:
            mock_subresource = mock.MagicMock()
            setattr(mock_rest_api, subresource, mock_subresource)

            for method in methods:
                mock_method = mock.MagicMock()
                setattr(mock_subresource, method, mock_method)

        return mock_rest_api

    @mock.patch(
        "manila.share.drivers.vastdata.rest.Session.refresh_auth_token"
    )
    def setUp(self, m_auth_token):
        super().setUp()
        self.fake_conf = configuration.Configuration(None)
        self._context = manila_context.get_admin_context()
        self._snapshot = fake_share.fake_snapshot_instance()

        self.fake_conf.set_default("driver_handles_share_servers", False)
        self.fake_conf.set_default("share_backend_name", "vast")
        self.fake_conf.set_default("vast_mgmt_host", "test")
        self.fake_conf.set_default("vast_root_export", "/fake")
        self.fake_conf.set_default("vast_vippool_name", "vippool")
        self.fake_conf.set_default("vast_mgmt_user", "user")
        self.fake_conf.set_default("vast_mgmt_password", "password")
        self._driver = driver.VASTShareDriver(
            execute=mock.MagicMock(), configuration=self.fake_conf
        )
        self._driver.do_setup(self._context)
        m_auth_token.assert_called_once()

    def test_do_setup(self):
        session = self._driver.rest.session
        self.assertEqual(self._driver._backend_name, "vast")
        self.assertEqual(self._driver._vippool_name, "vippool")
        self.assertEqual(self._driver._root_export, "/fake")
        self.assertFalse(session.ssl_verify)
        self.assertEqual(session.base_url, "https://test:443/api")

    @ddt.data("vast_mgmt_user", "vast_vippool_name")
    def test_do_setup_missing_required_fields(self, missing_field):
        self.fake_conf.set_default(missing_field, None)
        _driver = driver.VASTShareDriver(
            execute=mock.MagicMock(), configuration=self.fake_conf
        )
        with self.assertRaises(exception.VastDriverException):
            _driver.do_setup(self._context)

    @mock.patch(
        "manila.share.drivers.vastdata.rest.Session.get",
        mock.MagicMock(return_value=fake_metrics),
    )
    def test_update_share_stats(self):
        self._driver._update_share_stats()
        result = self._driver._stats
        self.assertEqual(result["share_backend_name"], "vast")
        self.assertEqual(result["driver_handles_share_servers"], False)
        self.assertEqual(result["vendor_name"], "VAST STORAGE")
        self.assertEqual(result["driver_version"], "1.0")
        self.assertEqual(result["storage_protocol"], "NFS")
        self.assertEqual(result["total_capacity_gb"], 471.1061706542969)
        self.assertEqual(result["free_capacity_gb"], 450.2256333641708)
        self.assertEqual(result["reserved_percentage"], 0)
        self.assertEqual(result["reserved_snapshot_percentage"], 0)
        self.assertEqual(result["reserved_share_extend_percentage"], 0)
        self.assertIs(result["qos"], False)
        self.assertIsNone(result["pools"])
        self.assertIs(result["snapshot_support"], True)
        self.assertIs(result["create_share_from_snapshot_support"], False)
        self.assertIs(result["revert_to_snapshot_support"], False)
        self.assertIs(result["mount_snapshot_support"], False)
        self.assertIsNone(result["replication_domain"])
        self.assertIsNone(result["filter_function"])
        self.assertIsNone(result["goodness_function"])
        self.assertIs(result["security_service_update_support"], False)
        self.assertIs(result["network_allocation_update_support"], False)
        self.assertIs(result["share_server_multiple_subnet_support"], False)
        self.assertIs(result["mount_point_name_support"], False)
        self.assertEqual(result["data_reduction"], 1.2)
        self.assertEqual(result["provisioned_capacity_gb"], 20.880537290126085)
        self.assertEqual(
            result["share_group_stats"],
            {"consistent_snapshot_support": None}
        )
        self.assertIs(result["ipv4_support"], True)
        self.assertIs(result["ipv6_support"], False)

    @ddt.idata(
        itertools.product(
            [1073741824, 1], ["NFS", "SMB"], ["fakeid", None]
        )
    )
    @ddt.unpack
    def test_create_shares(self, capacity, proto, policy):
        share = fake_share.fake_share(share_proto=proto)
        mock_rest = self._create_mocked_rest_api()
        mock_rest.view_policies.ensure.return_value = driver_util.Bunch(id=1)
        mock_rest.quotas.ensure.return_value = driver_util.Bunch(
            id=2, hard_limit=capacity
        )
        mock_rest.views.ensure.return_value = driver_util.Bunch(
            id=3, policy=policy
        )
        mock_rest.vip_pools.vips.return_value = ["1.1.1.0", "1.1.1.1"]
        with mock.patch.object(self._driver, "rest", mock_rest):
            if proto != "NFS":
                with self.assertRaises(exception.InvalidShare) as exc:
                    self._driver.create_share(self._context, share)
                self.assertIn(
                    "Invalid NAS protocol supplied",
                    str(exc.exception)
                )
            elif capacity == 1:
                with self.assertRaises(exception.ManilaException) as exc:
                    self._driver.create_share(self._context, share)
                self.assertIn(
                    "Share already exists with different capacity",
                    str(exc.exception)
                )
            else:

                location = self._driver.create_share(self._context, share)
                mock_rest.vip_pools.vips.assert_called_once_with(
                    pool_name="vippool"
                )
                mock_rest.view_policies.ensure.assert_called_once_with(
                    name="fakeid"
                )
                mock_rest.quotas.ensure.assert_called_once_with(
                    name="fakeid",
                    path="/fake/manila-fakeid",
                    create_dir=True,
                    hard_limit=capacity,
                )
                mock_rest.views.ensure.assert_called_once_with(
                    name="fakeid", path="/fake/manila-fakeid", policy_id=1
                )
                self.assertDictEqual(
                    location,
                    {
                        'path': '1.1.1.0:/fake/manila-fakeid',
                        'is_admin_only': False
                    }
                )
                if not policy:
                    mock_rest.views.update.assert_called_once_with(
                        3, policy_id=1
                    )
                else:
                    mock_rest.views.update.assert_not_called()

    def test_delete_share(self):
        share = fake_share.fake_share(share_proto="NFS")
        mock_rest = self._create_mocked_rest_api()
        with mock.patch.object(self._driver, "rest", mock_rest):
            self._driver.delete_share(self._context, share)
        mock_rest.folders.delete.assert_called_once_with(
            path="/fake/manila-fakeid"
        )
        mock_rest.views.delete.assert_called_once_with(name="fakeid")
        mock_rest.quotas.delete.assert_called_once_with(name="fakeid")
        mock_rest.view_policies.delete.assert_called_once_with(name="fakeid")

    def test_update_access_rules_wrong_proto(self):
        share = fake_share.fake_share(share_proto="SMB")
        access_rules = [
            {
                "access_level": "rw",
                "access_to": "127.0.0.1",
                "access_type": "ip"
            }
        ]
        res = self._driver.update_access(
            self._context,
            share,
            access_rules,
            None,
            None
        )
        self.assertIsNone(res)

    def test_update_access_add_rules_no_policy(self):
        share = fake_share.fake_share(share_proto="NFS")
        mock_rest = self._create_mocked_rest_api()
        mock_rest.view_policies.one.return_value = None
        access_rules = [
            {
                "access_level": "rw",
                "access_to": "127.0.0.1",
                "access_type": "ip"
            }
        ]
        with mock.patch.object(self._driver, "rest", mock_rest):
            with self.assertRaises(exception.ManilaException) as exc:
                self._driver.update_access(
                    self._context, share, access_rules, None, None
                )
            self.assertIn("Policy not found", str(exc.exception))

    @ddt.data(
        (["*"], ["10.10.10.1", "10.10.10.2"]),
        (["10.10.10.1", "10.10.10.2"], []),
        (["*"], []),
    )
    @ddt.unpack
    def test_update_access_add_rules(self, rw, ro):
        share = fake_share.fake_share(share_proto="NFS")
        mock_rest = self._create_mocked_rest_api()
        mock_rest.view_policies.one.return_value = driver_util.Bunch(
            id=1, nfs_read_write=rw, nfs_read_only=ro
        )
        access_rules = [
            {
                "access_level": "rw",
                "access_to": "127.0.0.1",
                "access_type": "ip"
            }
        ]
        with mock.patch.object(self._driver, "rest", mock_rest):
            failed_rules = self._driver.update_access(
                self._context,
                share,
                access_rules,
                None,
                None
            )

        expected_ro = set(ro)
        if rw == ["*"]:
            expected_rw = {"127.0.0.1"}
        else:
            expected_rw = set(["127.0.0.1"] + rw)
        kw = mock_rest.view_policies.update.call_args.kwargs
        self.assertEqual(kw["name"], "fakeid")
        self.assertSetEqual(set(kw["nfs_read_write"]), expected_rw)
        self.assertSetEqual(set(kw["nfs_read_only"]), expected_ro)
        self.assertEqual(kw["nfs_no_squash"], ["*"])
        self.assertEqual(kw["nfs_root_squash"], ["*"])
        self.assertFalse(failed_rules)

        # and the same for ro
        mock_rest = self._create_mocked_rest_api()
        mock_rest.view_policies.one.return_value = driver_util.Bunch(
            id=1, nfs_read_write=rw, nfs_read_only=ro
        )
        access_rules = [
            {
                "access_level": "ro",
                "access_to": "127.0.0.1",
                "access_type": "ip"
            }
        ]
        with mock.patch.object(self._driver, "rest", mock_rest):
            failed_rules = self._driver.update_access(
                self._context,
                share,
                access_rules,
                None,
                None
            )

        expected_rw = set(rw)
        if ro == ["*"]:
            expected_ro = {"127.0.0.1"}
        else:
            expected_ro = set(["127.0.0.1"] + ro)
        kw = mock_rest.view_policies.update.call_args.kwargs
        self.assertEqual(kw["name"], "fakeid")
        self.assertSetEqual(set(kw["nfs_read_write"]), expected_rw)
        self.assertSetEqual(set(kw["nfs_read_only"]), expected_ro)
        self.assertEqual(kw["nfs_no_squash"], ["*"])
        self.assertEqual(kw["nfs_root_squash"], ["*"])
        self.assertFalse(failed_rules)

    @ddt.data(
        (["*"], ["10.10.10.1", "10.10.10.2"]),
        (["10.10.10.1", "10.10.10.2"], []),
        (["*"], []),
    )
    @ddt.unpack
    def test_update_access_delete_rules(self, rw, ro):
        share = fake_share.fake_share(share_proto="NFS")
        mock_rest = self._create_mocked_rest_api()
        mock_rest.view_policies.one.return_value = driver_util.Bunch(
            id=1, nfs_read_write=rw, nfs_read_only=ro
        )
        delete_rules = [
            {
                "access_level": "rw",
                "access_to": "10.10.10.1",
                "access_type": "ip"
            }
        ]
        with mock.patch.object(self._driver, "rest", mock_rest):
            failed_rules = self._driver.update_access(
                self._context, share,
                None,
                None,
                delete_rules,
            )

        expected_ro = set(ro)
        if rw == ["*"]:
            expected_rw = set(rw)
        else:
            expected_rw = set([r for r in rw if r != "10.10.10.1"])
        kw = mock_rest.view_policies.update.call_args.kwargs
        self.assertEqual(kw["name"], "fakeid")
        self.assertSetEqual(set(kw["nfs_read_write"]), expected_rw)
        self.assertSetEqual(set(kw["nfs_read_only"]), expected_ro)
        self.assertEqual(kw["nfs_no_squash"], ["*"])
        self.assertEqual(kw["nfs_root_squash"], ["*"])
        self.assertFalse(failed_rules)

        # and the same for ro
        mock_rest = self._create_mocked_rest_api()
        mock_rest.view_policies.one.return_value = driver_util.Bunch(
            id=1, nfs_read_write=rw, nfs_read_only=ro
        )
        delete_rules = [
            {
                "access_level": "ro",
                "access_to": "10.10.10.1",
                "access_type": "ip"
            }
        ]
        with mock.patch.object(self._driver, "rest", mock_rest):
            failed_rules = self._driver.update_access(
                self._context, share, None, None, delete_rules
            )

        expected_rw = set(rw)
        if ro == ["*"]:
            expected_ro = set(ro)
        else:
            expected_ro = set([r for r in ro if r != "10.10.10.1"])
        kw = mock_rest.view_policies.update.call_args.kwargs
        self.assertEqual(kw["name"], "fakeid")
        self.assertSetEqual(set(kw["nfs_read_write"]), expected_rw)
        self.assertSetEqual(set(kw["nfs_read_only"]), expected_ro)
        self.assertEqual(kw["nfs_no_squash"], ["*"])
        self.assertEqual(kw["nfs_root_squash"], ["*"])
        self.assertFalse(failed_rules)

    def test_update_access_for_cidr(self):
        share = fake_share.fake_share(share_proto="NFS")
        mock_rest = self._create_mocked_rest_api()
        mock_rest.view_policies.one.return_value = driver_util.Bunch(
            id=1, nfs_read_write=["10.0.0.1"], nfs_read_only=["*"]
        )
        access_rules = [
            {
                "access_level": "ro",
                "access_to": "10.0.0.1/29",
                "access_type": "ip",
                "access_id": 12345,
            }
        ]
        with mock.patch.object(self._driver, "rest", mock_rest):
            failed_rules = self._driver.update_access(
                self._context,
                share,
                access_rules,
                None,
                None
            )
            kw = mock_rest.view_policies.update.call_args.kwargs
            self.assertEqual(kw["name"], "fakeid")
            self.assertSetEqual(set(kw["nfs_read_write"]), {"10.0.0.1"})
            self.assertSetEqual(
                set(kw["nfs_read_only"]),
                {
                    '10.0.0.1',
                    '10.0.0.3',
                    '10.0.0.2',
                    '10.0.0.6',
                    '10.0.0.5',
                    '10.0.0.4'
                }
            )
            self.assertFalse(failed_rules)

        delete_rules = [
            {
                "access_level": "ro",
                "access_to": "10.0.0.1/30",
                "access_type": "ip",
                "access_id": 12345,
            }
        ]
        mock_rest.view_policies.one.return_value = driver_util.Bunch(
            id=1, nfs_read_write=["10.0.0.1"],
            nfs_read_only=[
                '10.0.0.1',
                '10.0.0.3',
                '10.0.0.2',
                '10.0.0.6',
                '10.0.0.5',
                '10.0.0.4',
            ]
        )
        with mock.patch.object(self._driver, "rest", mock_rest):
            failed_rules = self._driver.update_access(
                self._context,
                share,
                None,
                None,
                delete_rules,
            )
            kw = mock_rest.view_policies.update.call_args.kwargs
            self.assertEqual(kw["name"], "fakeid")
            self.assertSetEqual(set(kw["nfs_read_write"]), {"10.0.0.1"})
            self.assertSetEqual(
                set(kw["nfs_read_only"]),
                {'10.0.0.6', '10.0.0.3', '10.0.0.4', '10.0.0.5'}
            )
            self.assertFalse(failed_rules)

    def test_update_access_for_invalid_rules(self):
        share = fake_share.fake_share(share_proto="NFS")
        mock_rest = self._create_mocked_rest_api()
        mock_rest.view_policies.one.return_value = driver_util.Bunch(
            id=1, nfs_read_write=["10.0.0.1"], nfs_read_only=["*"]
        )
        access_rules = [
            {
                "access_level": "ry",
                "access_to": "10.0.0.1",
                "access_type": "ip",
                "access_id": 12345,
                "id": 12345,
            },
            {
                "access_level": "ro",
                "access_to": "10.0.0.2",
                "access_type": "ip",
                "access_id": 12346,
                "id": 12346,
            },
            {
                "access_level": "ro",
                "access_to": "10.0.0.2/33",
                "access_type": "ip",
                "access_id": 12347,
                "id": 12347,
            },
            {
                "access_level": "rw",
                "access_to": "10.0.0.2.4",
                "access_type": "ip",
                "access_id": 12348,
                "id": 12348,
            }
        ]
        with mock.patch.object(self._driver, "rest", mock_rest):
            failed_rules = self._driver.update_access(
                self._context,
                share,
                access_rules,
                None,
                None
            )
            kw = mock_rest.view_policies.update.call_args.kwargs
            self.assertEqual(kw["name"], "fakeid")
            self.assertSetEqual(set(kw["nfs_read_write"]), {"10.0.0.1"})
            self.assertSetEqual(set(kw["nfs_read_only"]), {'10.0.0.2'})
            self.assertDictEqual(
                failed_rules,
                {
                    12345: {'state': 'error'},
                    12347: {'state': 'error'},
                    12348: {'state': 'error'}
                }
            )

    def test_resize_share_quota_not_found(self):
        share = fake_share.fake_share(share_proto="NFS")
        mock_rest = self._create_mocked_rest_api()
        mock_rest.quotas.one.return_value = None
        with mock.patch.object(self._driver, "rest", mock_rest):
            with self.assertRaises(exception.ShareNotFound) as exc:
                self._driver.extend_share(share, 10000)
            self.assertIn("could not be found", str(exc.exception))

    def test_resize_share_ok(self):
        share = fake_share.fake_share(share_proto="NFS")
        mock_rest = self._create_mocked_rest_api()
        mock_rest.quotas.one.return_value = driver_util.Bunch(
            id=1, used_effective_capacity=1073741824
        )
        with mock.patch.object(self._driver, "rest", mock_rest):
            self._driver.extend_share(share, 50)
            mock_rest.quotas.update.assert_called_with(
                1, hard_limit=53687091200
            )
            mock_rest.quotas.update.reset()
            self._driver.shrink_share(share, 20)
            mock_rest.quotas.update.assert_called_with(
                1, hard_limit=21474836480
            )

    def test_resize_share_exceeded_hard_limit(self):
        share = fake_share.fake_share(
            share_proto="NFS"
        )
        mock_rest = self._create_mocked_rest_api()
        mock_rest.quotas.one.return_value = driver_util.Bunch(
            id=1, used_effective_capacity=10737418240
        )  # 10GB
        with mock.patch.object(self._driver, "rest", mock_rest):
            with self.assertRaises(exception.ShareShrinkingPossibleDataLoss):
                self._driver.shrink_share(share, 9.7)
            self._driver.shrink_share(share, 10)

    def test_create_snapshot(self):
        snapshot = driver_util.Bunch(
            name="fakesnap", share_instance_id="fakeid"
        )
        mock_rest = self._create_mocked_rest_api()
        with mock.patch.object(self._driver, "rest", mock_rest):
            self._driver.create_snapshot(self._context, snapshot, None)
        mock_rest.snapshots.create.assert_called_once_with(
            path="/fake/manila-fakeid", name="fakesnap"
        )

    def test_delete_snapshot(self):
        snapshot = driver_util.Bunch(
            name="fakesnap", share_instance_id="fakeid"
        )
        mock_rest = self._create_mocked_rest_api()
        with mock.patch.object(self._driver, "rest", mock_rest):
            self._driver.delete_snapshot(self._context, snapshot, None)
        mock_rest.snapshots.delete.assert_called_once_with(name="fakesnap")

    def test_network_allocation_number(self):
        self.assertEqual(self._driver.get_network_allocations_number(), 0)

    @ddt.data([], ['fake/path/1', 'fake/path'])
    def test_ensure_shares(self, fake_export_locations):
        mock_rest = self._create_mocked_rest_api()
        mock_rest.view_policies.ensure.return_value = driver_util.Bunch(id=1)
        mock_rest.quotas.ensure.return_value = driver_util.Bunch(
            id=2, hard_limit=1073741824
        )
        mock_rest.views.ensure.return_value = driver_util.Bunch(
            id=3, policy="test_policy"
        )
        shares = [
            fake_share.fake_share(
                id=_id,
                share_id=share_id,
                share_proto="NFS",
                export_locations=fake_export_locations,
            )
            for _id, share_id in enumerate(["123", "456", "789"], 1)
        ]
        mock_rest.vip_pools.vips.return_value = ["1.1.1.0", "1.1.1.1"]
        with mock.patch.object(self._driver, "rest", mock_rest):
            locations = self._driver.ensure_shares(self._context, shares)

        common = {"is_admin_only": False}
        self.assertDictEqual(
            locations,
            {
                1: {
                    "export_locations": [
                        {"path": "1.1.1.0:/fake/manila-1", **common},
                        {"path": "1.1.1.1:/fake/manila-1", **common},
                    ]
                },
                2: {
                    "export_locations": [
                        {"path": "1.1.1.0:/fake/manila-2", **common},
                        {"path": "1.1.1.1:/fake/manila-2", **common},
                    ]
                },
                3: {
                    "export_locations": [
                        {"path": "1.1.1.0:/fake/manila-3", **common},
                        {"path": "1.1.1.1:/fake/manila-3", **common},
                    ]
                },
            },
        )

    def test_backend_info(self):
        backend_info = self._driver.get_backend_info(self._context)
        self.assertDictEqual(
            backend_info,
            {'vast_vippool_name': 'vippool', 'vast_mgmt_host': 'test'}
        )


class TestPolicyPayloadFromRules(unittest.TestCase):
    def test_policy_payload_from_rules_update(self):
        rules = [{"access_level": "rw", "access_to": "127.0.0.1"}]
        policy = mock.MagicMock()
        policy.nfs_read_write = ["127.0.0.1"]
        policy.nfs_read_only = []
        result = driver.policy_payload_from_rules(rules, policy, "update")
        self.assertEqual(
            result, {"nfs_read_write": ["127.0.0.1"], "nfs_read_only": []}
        )

    def test_policy_payload_from_rules_deny(self):
        rules = [{"access_level": "rw", "access_to": "127.0.0.1"}]
        policy = mock.MagicMock()
        policy.nfs_read_write = ["127.0.0.1"]
        policy.nfs_read_only = []
        result = driver.policy_payload_from_rules(rules, policy, "deny")
        self.assertEqual(result, {"nfs_read_write": [], "nfs_read_only": []})

    def test_policy_payload_from_rules_invalid_action(self):
        rules = [{"access_level": "rw", "access_to": "127.0.0.1"}]
        with self.assertRaises(ValueError):
            driver.policy_payload_from_rules(rules, None, "invalid")

    def test_policy_payload_from_rules_invalid_ip(self):
        rules = [{"access_level": "rw", "access_to": "1.0.0.257"}]
        with self.assertRaises(netaddr.core.AddrFormatError):
            driver.policy_payload_from_rules(rules, None, "deny")


class TestValidateAccessRules(unittest.TestCase):
    def test_validate_access_rules_invalid_type(self):
        rule = {"access_type": "INVALID", "access_level": "rw"}
        with self.assertRaises(exception.InvalidShareAccess):
            driver.validate_access_rule(rule)

    def test_validate_access_rules_invalid_level(self):
        rule = {"access_type": "ip", "access_level": "INVALID"}
        with self.assertRaises(exception.InvalidShareAccessLevel):
            driver.validate_access_rule(rule)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import versionutils
from oslo_policy import policy

from manila.policies import base


BASE_POLICY_NAME = 'security_service:%s'

DEPRECATED_REASON = """
The security service API now supports scope and default roles.
"""

deprecated_security_service_create = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'create',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_security_service_show = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'show',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_security_service_detail = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'detail',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_security_service_index = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'index',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_security_service_update = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'update',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_security_service_delete = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'delete',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_security_service_get_all = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'get_all_security_services',
    check_str=base.RULE_ADMIN_API,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)


security_service_policies = [
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'create',
        check_str=base.ADMIN_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description="Create security service.",
        operations=[
            {
                'method': 'POST',
                'path': '/security-services'
            }
        ],
        deprecated_rule=deprecated_security_service_create
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'show',
        check_str=base.ADMIN_OR_PROJECT_READER,
        scope_types=['project'],
        description="Get details of a security service.",
        operations=[
            {
                'method': 'GET',
                'path': '/security-services/{security_service_id}'
            }
        ],
        deprecated_rule=deprecated_security_service_show
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'detail',
        check_str=base.ADMIN_OR_PROJECT_READER,
        scope_types=['project'],
        description="Get details of all security services.",
        operations=[
            {
                'method': 'GET',
                'path': '/security-services/detail?{query}'
            },
        ],
        deprecated_rule=deprecated_security_service_detail
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'index',
        check_str=base.ADMIN_OR_PROJECT_READER,
        scope_types=['project'],
        description="Get all security services under a project.",
        operations=[
            {
                'method': 'GET',
                'path': '/security-services?{query}'
            }
        ],
        deprecated_rule=deprecated_security_service_index
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'update',
        check_str=base.ADMIN_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description="Update a security service.",
        operations=[
            {
                'method': 'PUT',
                'path': '/security-services/{security_service_id}',
            }
        ],
        deprecated_rule=deprecated_security_service_update
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'delete',
        check_str=base.ADMIN_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description="Delete a security service.",
        operations=[
            {
                'method': 'DELETE',
                'path': '/security-services/{security_service_id}'
            }
        ],
        deprecated_rule=deprecated_security_service_delete
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'get_all_security_services',
        check_str=base.ADMIN,
        scope_types=['project'],
        description="Get security services of all projects.",
        operations=[
            {
                'method': 'GET',
                'path': '/security-services?all_tenants=1'
            },
            {
                'method': 'GET',
                'path': '/security-services/detail?all_tenants=1'
            }
        ],
        deprecated_rule=deprecated_security_service_get_all
    ),
]


def list_rules():
    return security_service_policies



# Copyright (C) 2017 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from http import client as http
import re

from manila.api.openstack import api_version_request as api_version
from manila.api import validation
from manila.api.validation import parameter_types
from manila import exception
from manila import test


class FakeRequest(object):
    api_version_request = api_version.APIVersionRequest("3.0")
    environ = {}


class APIValidationTestCase(test.TestCase):
    def setUp(self, schema=None):
        super().setUp()
        self.post = None

        if schema is not None:

            @validation.request_body_schema(schema=schema)
            def post(req, body):
                return 'Validation succeeded.'

            self.post = post

    def check_validation_error(self, method, body, expected_detail, req=None):
        if not req:
            req = FakeRequest()
        try:
            method(
                body=body,
                req=req,
            )
        except exception.ValidationError as ex:
            self.assertEqual(http.BAD_REQUEST, ex.kwargs['code'])
            if isinstance(expected_detail, list):
                self.assertIn(
                    ex.kwargs['detail'],
                    expected_detail,
                    'Exception details did not match expected',
                )
            elif not re.match(expected_detail, ex.kwargs['detail']):
                self.assertEqual(
                    expected_detail,
                    ex.kwargs['detail'],
                    'Exception details did not match expected',
                )
        except Exception as ex:
            self.fail('An unexpected exception happens: %s' % ex)
        else:
            self.fail('Any exception did not happen.')


class RequiredDisableTestCase(APIValidationTestCase):
    def setUp(self):
        schema = {
            'type': 'object',
            'properties': {
                'foo': {
                    'type': 'integer',
                },
            },
        }
        super().setUp(schema=schema)

    def test_validate_required_disable(self):
        self.assertEqual(
            'Validation succeeded.',
            self.post(body={'foo': 1}, req=FakeRequest()),
        )


class RequiredEnableTestCase(APIValidationTestCase):
    def setUp(self):
        schema = {
            'type': 'object',
            'properties': {
                'foo': {
                    'type': 'integer',
                },
            },
            'required': ['foo'],
        }
        super().setUp(schema=schema)

    def test_validate_required_enable(self):
        self.assertEqual(
            'Validation succeeded.',
            self.post(body={'foo': 1}, req=FakeRequest()),
        )

    def test_validate_required_enable_fails(self):
        detail = "'foo' is a required property"
        self.check_validation_error(
            self.post, body={'abc': 1}, expected_detail=detail
        )


class AdditionalPropertiesEnableTestCase(APIValidationTestCase):
    def setUp(self):
        schema = {
            'type': 'object',
            'properties': {
                'foo': {
                    'type': 'integer',
                },
            },
            'required': ['foo'],
        }
        super().setUp(schema=schema)

    def test_validate_additionalProperties_enable(self):
        self.assertEqual(
            'Validation succeeded.',
            self.post(body={'foo': 1}, req=FakeRequest()),
        )
        self.assertEqual(
            'Validation succeeded.',
            self.post(body={'foo': 1, 'ext': 1}, req=FakeRequest()),
        )


class AdditionalPropertiesDisableTestCase(APIValidationTestCase):
    def setUp(self):
        schema = {
            'type': 'object',
            'properties': {
                'foo': {
                    'type': 'integer',
                },
            },
            'required': ['foo'],
            'additionalProperties': False,
        }
        super().setUp(schema=schema)

    def test_validate_additionalProperties_disable(self):
        self.assertEqual(
            'Validation succeeded.',
            self.post(body={'foo': 1}, req=FakeRequest()),
        )

    def test_validate_additionalProperties_disable_fails(self):
        detail = "Additional properties are not allowed ('ext' was unexpected)"
        self.check_validation_error(
            self.post, body={'foo': 1, 'ext': 1}, expected_detail=detail
        )


class PatternPropertiesTestCase(APIValidationTestCase):
    def setUp(self):
        schema = {
            'patternProperties': {
                '^[a-zA-Z0-9]{1,10}$': {'type': 'string'},
            },
            'additionalProperties': False,
        }
        super().setUp(schema=schema)

    def test_validate_patternProperties(self):
        self.assertEqual(
            'Validation succeeded.',
            self.post(body={'foo': 'bar'}, req=FakeRequest()),
        )

    def test_validate_patternProperties_fails(self):
        details = [
            "Additional properties are not allowed ('__' was unexpected)",
            "'__' does not match any of the regexes: '^[a-zA-Z0-9]{1,10}$'",
        ]
        self.check_validation_error(
            self.post, body={'__': 'bar'}, expected_detail=details
        )

        details = [
            "'' does not match any of the regexes: '^[a-zA-Z0-9]{1,10}$'",
            "Additional properties are not allowed ('' was unexpected)",
        ]
        self.check_validation_error(
            self.post, body={'': 'bar'}, expected_detail=details
        )

        details = [
            (
                "'0123456789a' does not match any of the regexes: "
                "'^[a-zA-Z0-9]{1,10}$'"
            ),
            (
                "Additional properties are not allowed ('0123456789a' was "
                "unexpected)"
            ),
        ]
        self.check_validation_error(
            self.post, body={'0123456789a': 'bar'}, expected_detail=details
        )

        detail = "expected string or bytes-like object"
        self.check_validation_error(
            self.post, body={None: 'bar'}, expected_detail=detail
        )


class StringTestCase(APIValidationTestCase):
    def setUp(self):
        schema = {
            'type': 'object',
            'properties': {
                'foo': {
                    'type': 'string',
                },
            },
        }
        super().setUp(schema=schema)

    def test_validate_string(self):
        self.assertEqual(
            'Validation succeeded.',
            self.post(body={'foo': 'abc'}, req=FakeRequest()),
        )
        self.assertEqual(
            'Validation succeeded.',
            self.post(body={'foo': '0'}, req=FakeRequest()),
        )
        self.assertEqual(
            'Validation succeeded.',
            self.post(body={'foo': ''}, req=FakeRequest()),
        )

    def test_validate_string_fails(self):
        detail = (
            "Invalid input for field/attribute foo. Value: 1. "
            "1 is not of type 'string'"
        )
        self.check_validation_error(
            self.post, body={'foo': 1}, expected_detail=detail
        )

        detail = (
            "Invalid input for field/attribute foo. Value: 1.5. "
            "1.5 is not of type 'string'"
        )
        self.check_validation_error(
            self.post, body={'foo': 1.5}, expected_detail=detail
        )

        detail = (
            "Invalid input for field/attribute foo. Value: True. "
            "True is not of type 'string'"
        )
        self.check_validation_error(
            self.post, body={'foo': True}, expected_detail=detail
        )


class StringLengthTestCase(APIValidationTestCase):
    def setUp(self):
        schema = {
            'type': 'object',
            'properties': {
                'foo': {
                    'type': 'string',
                    'minLength': 1,
                    'maxLength': 10,
                },
            },
        }
        super().setUp(schema=schema)

    def test_validate_string_length(self):
        self.assertEqual(
            'Validation succeeded.',
            self.post(body={'foo': '0'}, req=FakeRequest()),
        )
        self.assertEqual(
            'Validation succeeded.',
            self.post(body={'foo': '0123456789'}, req=FakeRequest()),
        )

    def test_validate_string_length_fails(self):
        # checks for jsonschema output from 3.2.x and 4.21.x
        detail = (
            "Invalid input for field/attribute foo. Value: . "
            "'' (is too short|should be non-empty)"
        )
        self.check_validation_error(
            self.post, body={'foo': ''}, expected_detail=detail
        )

        detail = (
            "Invalid input for field/attribute foo. Value: 0123456789a. "
            "'0123456789a' is too long"
        )
        self.check_validation_error(
            self.post, body={'foo': '0123456789a'}, expected_detail=detail
        )


class IntegerTestCase(APIValidationTestCase):
    def setUp(self):
        schema = {
            'type': 'object',
            'properties': {
                'foo': {
                    'type': ['integer', 'string'],
                    'pattern': '^[0-9]+$',
                },
            },
        }
        super().setUp(schema=schema)

    def test_validate_integer(self):
        self.assertEqual(
            'Validation succeeded.',
            self.post(body={'foo': 1}, req=FakeRequest()),
        )
        self.assertEqual(
            'Validation succeeded.',
            self.post(body={'foo': '1'}, req=FakeRequest()),
        )
        self.assertEqual(
            'Validation succeeded.',
            self.post(body={'foo': '0123456789'}, req=FakeRequest()),
        )

    def test_validate_integer_fails(self):
        detail = (
            "Invalid input for field/attribute foo. Value: abc. "
            "'abc' does not match '^[0-9]+$'"
        )
        self.check_validation_error(
            self.post, body={'foo': 'abc'}, expected_detail=detail
        )

        detail = (
            "Invalid input for field/attribute foo. Value: True. "
            "True is not of type 'integer', 'string'"
        )
        self.check_validation_error(
            self.post, body={'foo': True}, expected_detail=detail
        )

        detail = (
            "Invalid input for field/attribute foo. Value: 0xffff. "
            "'0xffff' does not match '^[0-9]+$'"
        )
        self.check_validation_error(
            self.post, body={'foo': '0xffff'}, expected_detail=detail
        )

        detail = (
            "Invalid input for field/attribute foo. Value: 1.01. "
            "1.01 is not of type 'integer', 'string'"
        )
        self.check_validation_error(
            self.post, body={'foo': 1.01}, expected_detail=detail
        )

        detail = (
            "Invalid input for field/attribute foo. Value: 1.0. "
            "'1.0' does not match '^[0-9]+$'"
        )
        self.check_validation_error(
            self.post, body={'foo': '1.0'}, expected_detail=detail
        )


class IntegerRangeTestCase(APIValidationTestCase):
    def setUp(self):
        schema = {
            'type': 'object',
            'properties': {
                'foo': {
                    'type': ['integer', 'string'],
                    'pattern': '^[0-9]+$',
                    'minimum': 1,
                    'maximum': 10,
                },
            },
        }
        super().setUp(schema=schema)

    def test_validate_integer_range(self):
        self.assertEqual(
            'Validation succeeded.',
            self.post(body={'foo': 1}, req=FakeRequest()),
        )
        self.assertEqual(
            'Validation succeeded.',
            self.post(body={'foo': 10}, req=FakeRequest()),
        )
        self.assertEqual(
            'Validation succeeded.',
            self.post(body={'foo': '1'}, req=FakeRequest()),
        )

    def test_validate_integer_range_fails(self):
        detail = (
            "Invalid input for field/attribute foo. Value: 0. "
            "0(.0)? is less than the minimum of 1"
        )
        self.check_validation_error(
            self.post, body={'foo': 0}, expected_detail=detail
        )

        detail = (
            "Invalid input for field/attribute foo. Value: 11. "
            "11(.0)? is greater than the maximum of 10"
        )
        self.check_validation_error(
            self.post, body={'foo': 11}, expected_detail=detail
        )

        detail = (
            "Invalid input for field/attribute foo. Value: 0. "
            "0(.0)? is less than the minimum of 1"
        )
        self.check_validation_error(
            self.post, body={'foo': '0'}, expected_detail=detail
        )

        detail = (
            "Invalid input for field/attribute foo. Value: 11. "
            "11(.0)? is greater than the maximum of 10"
        )
        self.check_validation_error(
            self.post, body={'foo': '11'}, expected_detail=detail
        )


class BooleanTestCase(APIValidationTestCase):
    def setUp(self):
        schema = {
            'type': 'object',
            'properties': {
                'foo': parameter_types.boolean,
            },
        }
        super().setUp(schema=schema)

    def test_validate_boolean(self):
        self.assertEqual(
            'Validation succeeded.',
            self.post(body={'foo': True}, req=FakeRequest()),
        )
        self.assertEqual(
            'Validation succeeded.',
            self.post(body={'foo': False}, req=FakeRequest()),
        )
        self.assertEqual(
            'Validation succeeded.',
            self.post(body={'foo': 'True'}, req=FakeRequest()),
        )
        self.assertEqual(
            'Validation succeeded.',
            self.post(body={'foo': 'False'}, req=FakeRequest()),
        )
        self.assertEqual(
            'Validation succeeded.',
            self.post(body={'foo': '1'}, req=FakeRequest()),
        )
        self.assertEqual(
            'Validation succeeded.',
            self.post(body={'foo': '0'}, req=FakeRequest()),
        )

    def test_validate_boolean_fails(self):
        enum_boolean = (
            "[True, 'True', 'TRUE', 'true', '1', 'ON', 'On', "
            "'on', 'YES', 'Yes', 'yes', 'y', 't', "
            "False, 'False', 'FALSE', 'false', '0', 'OFF', 'Off', "
            "'off', 'NO', 'No', 'no', 'n', 'f']"
        )

        detail = (
            "Invalid input for field/attribute foo. Value: bar. "
            "'bar' is not one of %s"
        ) % enum_boolean
        self.check_validation_error(
            self.post, body={'foo': 'bar'}, expected_detail=detail
        )

        detail = (
            "Invalid input for field/attribute foo. Value: 2. "
            "'2' is not one of %s"
        ) % enum_boolean
        self.check_validation_error(
            self.post, body={'foo': '2'}, expected_detail=detail
        )


class DatetimeTestCase(APIValidationTestCase):
    def setUp(self):
        schema = {
            'type': 'object',
            'properties': {
                'foo': {
                    'type': ['string', 'null'],
                    'format': 'date-time',
                },
            },
        }
        super().setUp(schema=schema)

    def test_validate_datetime(self):
        self.assertEqual(
            'Validation succeeded.',
            self.post(body={'foo': '2017-01-14T01:00:00Z'}, req=FakeRequest()),
        )
        self.assertEqual(
            'Validation succeeded.',
            self.post(body={'foo': None}, req=FakeRequest()),
        )

    def test_validate_datetime_fails(self):
        detail = (
            "Invalid input for field/attribute foo. Value: True. "
            "True is not of type 'string', 'null'"
        )
        self.check_validation_error(
            self.post, body={'foo': True}, expected_detail=detail
        )

        detail = (
            "Invalid input for field/attribute foo. Value: 123. "
            "'123' is not a 'date-time'"
        )
        self.check_validation_error(
            self.post, body={'foo': '123'}, expected_detail=detail
        )



#    Copyright 2023 Cloudification GmbH.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from manila.data import backup_driver


class FakeBackupDriver(backup_driver.BackupDriver):
    """Fake Backup driver."""

    def __init__(self, *args, **kwargs):
        super(FakeBackupDriver, self).__init__(*args, **kwargs)
        pass

    def backup(self, backup, share):
        """Start a backup of a specified share."""
        pass

    def restore(self, backup, share):
        """Restore a saved backup."""
        pass

    def delete(self, backup):
        """Delete a saved backup."""
        pass

    def get_backup_info(self, backup):
        """Get backup capabilities information of driver."""
        backup_info = {
            'mount': 'mount -vt fake_proto /fake-export %(path)s',
            'unmount': 'umount -v %(path)s',
        }
        return backup_info



# Copyright 2015 Mirantis inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log
from webob import exc

from manila.api.openstack import wsgi
from manila.api.v2 import metadata
from manila.api.views import export_locations as export_locations_views
from manila.db import api as db_api
from manila import exception
from manila.i18n import _
from manila import policy

LOG = log.getLogger(__name__)
CONF = cfg.CONF


class ShareExportLocationController(wsgi.Controller,
                                    metadata.MetadataController):
    """The Share Export Locations API controller."""

    def __init__(self):
        self._view_builder_class = export_locations_views.ViewBuilder
        self.resource_name = 'share_export_location'
        super(ShareExportLocationController, self).__init__()
        self._conf_admin_only_metadata_keys = getattr(
            CONF, 'admin_only_el_metadata', []
        )

    def _verify_share(self, context, share_id):
        try:
            share = db_api.share_get(context, share_id)
            if not share['is_public']:
                policy.check_policy(context, 'share', 'get', share)
        except exception.NotFound:
            msg = _("Share '%s' not found.") % share_id
            raise exc.HTTPNotFound(explanation=msg)

    @wsgi.Controller.authorize('index')
    def _index(self, req, share_id, ignore_secondary_replicas=False):
        context = req.environ['manila.context']
        self._verify_share(context, share_id)
        kwargs = {
            'include_admin_only': context.is_admin,
            'ignore_migration_destination': True,
            'ignore_secondary_replicas': ignore_secondary_replicas,
        }
        export_locations = db_api.export_location_get_all_by_share_id(
            context, share_id, **kwargs)
        return self._view_builder.summary_list(req, export_locations)

    @wsgi.Controller.authorize('show')
    def _show(self, req, share_id, export_location_uuid,
              ignore_secondary_replicas=False):
        context = req.environ['manila.context']
        self._verify_share(context, share_id)
        try:
            export_location = db_api.export_location_get_by_uuid(
                context, export_location_uuid,
                ignore_secondary_replicas=ignore_secondary_replicas)
        except exception.ExportLocationNotFound:
            msg = _("Export location '%s' not found.") % export_location_uuid
            raise exc.HTTPNotFound(explanation=msg)

        if export_location.is_admin_only and not context.is_admin:
            raise exc.HTTPForbidden()

        return self._view_builder.detail(req, export_location)

    @wsgi.Controller.api_version('2.9', '2.46')
    def index(self, req, share_id):
        """Return a list of export locations for share."""
        return self._index(req, share_id)

    @wsgi.Controller.api_version('2.47')  # noqa: F811
    def index(self, req, share_id):  # pylint: disable=function-redefined  # noqa F811
        """Return a list of export locations for share."""
        return self._index(req, share_id,
                           ignore_secondary_replicas=True)

    @wsgi.Controller.api_version('2.9', '2.46')
    def show(self, req, share_id, export_location_uuid):
        """Return data about the requested export location."""
        return self._show(req, share_id, export_location_uuid)

    @wsgi.Controller.api_version('2.47')  # noqa: F811
    def show(self, req, share_id,  # pylint: disable=function-redefined  # noqa F811
             export_location_uuid):
        """Return data about the requested export location."""
        return self._show(req, share_id, export_location_uuid,
                          ignore_secondary_replicas=True)

    def _validate_metadata_for_update(self, req, share_export_location,
                                      metadata, delete=True):
        persistent_keys = set(self._conf_admin_only_metadata_keys)
        context = req.environ['manila.context']
        if set(metadata).intersection(persistent_keys):
            try:
                policy.check_policy(
                    context, 'share_export_location',
                    'update_admin_only_metadata')
            except exception.PolicyNotAuthorized:
                msg = _("Cannot set or update admin only metadata.")
                LOG.exception(msg)
                raise exc.HTTPForbidden(explanation=msg)
            persistent_keys = []

        current_export_metadata = db_api.export_location_metadata_get(
            context, share_export_location)
        if delete:
            _metadata = metadata
            for key in persistent_keys:
                if key in current_export_metadata:
                    _metadata[key] = current_export_metadata[key]
        else:
            metadata_copy = metadata.copy()
            for key in persistent_keys:
                metadata_copy.pop(key, None)
            _metadata = current_export_metadata.copy()
            _metadata.update(metadata_copy)

        return _metadata

    @wsgi.Controller.api_version("2.87")
    @wsgi.Controller.authorize("get_metadata")
    def index_metadata(self, req, share_id, resource_id):
        """Returns the list of metadata for a given share export location."""
        context = req.environ['manila.context']
        self._verify_share(context, share_id)
        return self._index_metadata(req, resource_id)

    @wsgi.Controller.api_version("2.87")
    @wsgi.Controller.authorize("update_metadata")
    def create_metadata(self, req, share_id, resource_id, body):
        """Create metadata for a given share export location."""
        _metadata = self._validate_metadata_for_update(req, resource_id,
                                                       body['metadata'],
                                                       delete=False)
        body['metadata'] = _metadata
        context = req.environ['manila.context']
        self._verify_share(context, share_id)
        return self._create_metadata(req, resource_id, body)

    @wsgi.Controller.api_version("2.87")
    @wsgi.Controller.authorize("update_metadata")
    def update_all_metadata(self, req, share_id, resource_id, body):
        """Update entire metadata for a given share export location."""
        _metadata = self._validate_metadata_for_update(req, resource_id,
                                                       body['metadata'])
        body['metadata'] = _metadata
        context = req.environ['manila.context']
        self._verify_share(context, share_id)
        return self._update_all_metadata(req, resource_id, body)

    @wsgi.Controller.api_version("2.87")
    @wsgi.Controller.authorize("update_metadata")
    def update_metadata_item(self, req, share_id, resource_id, body, key):
        """Update metadata item for a given share export location."""
        _metadata = self._validate_metadata_for_update(req, resource_id,
                                                       body['metadata'],
                                                       delete=False)
        body['metadata'] = _metadata
        context = req.environ['manila.context']
        self._verify_share(context, share_id)
        return self._update_metadata_item(req, resource_id, body, key)

    @wsgi.Controller.api_version("2.87")
    @wsgi.Controller.authorize("get_metadata")
    def show_metadata(self, req, share_id, resource_id, key):
        """Show metadata for a given share export location."""
        context = req.environ['manila.context']
        self._verify_share(context, share_id)
        return self._show_metadata(req, resource_id, key)

    @wsgi.Controller.api_version("2.87")
    @wsgi.Controller.authorize("delete_metadata")
    def delete_metadata(self, req, share_id, resource_id, key):
        """Delete metadata for a given share export location."""
        context = req.environ['manila.context']
        self._verify_share(context, share_id)
        return self._delete_metadata(req, resource_id, key)


def create_resource():
    return wsgi.Resource(ShareExportLocationController())



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Common parameter types for validating API requests."""

from manila.common import constants


def single_param(schema):
    """Macro function to support query params that allow only one value."""
    ret = multi_params(schema)
    ret['maxItems'] = 1
    return ret


def multi_params(schema):
    """Macro function to support query params that allow multiple values."""
    return {'type': 'array', 'items': schema}


boolean = {
    'type': ['boolean', 'string'],
    'enum': [
        True,
        'True',
        'TRUE',
        'true',
        '1',
        'ON',
        'On',
        'on',
        'YES',
        'Yes',
        'yes',
        'y',
        't',
        False,
        'False',
        'FALSE',
        'false',
        '0',
        'OFF',
        'Off',
        'off',
        'NO',
        'No',
        'no',
        'n',
        'f',
    ],
}

positive_integer = {
    'type': ['integer', 'string'],
    'pattern': '^[0-9]*$',
    'minimum': 1,
    'maximum': constants.DB_MAX_INT,
    'minLength': 1,
}

non_negative_integer = {
    'type': ['integer', 'string'],
    'pattern': '^[0-9]*$',
    'minimum': 0,
    'maximum': constants.DB_MAX_INT,
    'minLength': 1,
}



# Copyright (c) 2023 EMC Corporation.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

import ddt
from oslo_log import log
from oslo_utils import units

from manila.common import constants as const
from manila import exception
from manila.share.drivers.dell_emc.plugins.powerflex import connection
from manila import test

LOG = log.getLogger(__name__)


@ddt.ddt
class PowerFlexTest(test.TestCase):
    """Integration test for the PowerFlex Manila driver."""

    POWERFLEX_ADDR = "192.168.0.110"
    SHARE_NAME = "Manila-UT-filesystem"
    STORAGE_POOL_ID = "28515fee00000000"
    FILESYSTEM_ID = "6432b79e-1cc3-0414-3ffd-2a50fb1ccff3"
    NFS_EXPORT_ID = "6433a2b2-6d60-f737-9f3b-2a50fb1ccff3"
    NFS_EXPORT_NAME = "Manila-UT-filesystem"
    SNAPSHOT_NAME = "Manila-UT-filesystem-snap"
    SNAPSHOT_PATH = "Manila-UT-filesystem"
    SNAPSHOT_ID = "75758d63-2946-4c07-9118-9a6c6027d5e7"
    NAS_SERVER_IP = "192.168.11.23"

    class MockConfig(object):
        def safe_get(self, value):
            if value == "dell_nas_backend_host":
                return "192.168.0.110"
            elif value == "dell_nas_backend_port":
                return 443
            elif value == "dell_nas_login":
                return "admin"
            elif value == "dell_nas_password":
                return "pwd"
            elif value == "powerflex_storage_pool":
                return "Env8-SP-SW_SSD-1"
            elif value == "powerflex_protection_domain":
                return "Env8-PD-1"
            elif value == "dell_nas_server":
                return "env8nasserver"
            else:
                return None

    @mock.patch(
        "manila.share.drivers.dell_emc.plugins.powerflex.object_manager."
        "StorageObjectManager",
        autospec=True,
    )
    def setUp(self, mock_powerflex_manager):
        super(PowerFlexTest, self).setUp()

        self._mock_powerflex_manager = mock_powerflex_manager.return_value
        self.storage_connection = connection.PowerFlexStorageConnection(LOG)

        self.mock_context = mock.Mock("Context")
        self.mock_emc_driver = mock.Mock("EmcDriver")

        self._mock_config = self.MockConfig()
        self.mock_emc_driver.attach_mock(self._mock_config, "configuration")
        self.storage_connection.connect(
            self.mock_emc_driver, self.mock_context
        )

    @mock.patch(
        "manila.share.drivers.dell_emc.plugins.powerflex.object_manager."
        "StorageObjectManager",
        autospec=True,
    )
    def test_connect(self, mock_powerflex_manager):
        storage_connection = connection.PowerFlexStorageConnection(LOG)

        # execute method under test
        storage_connection.connect(self.mock_emc_driver, self.mock_context)

        # verify connect sets driver params appropriately
        mock_config = self.MockConfig()
        server_addr = mock_config.safe_get("dell_nas_backend_host")
        self.assertEqual(server_addr, storage_connection.rest_ip)
        expected_port = int(mock_config.safe_get("dell_nas_backend_port"))
        self.assertEqual(expected_port, storage_connection.rest_port)
        self.assertEqual(
            "https://{0}:{1}".format(server_addr, expected_port),
            storage_connection.host_url,
        )
        expected_username = mock_config.safe_get("dell_nas_login")
        self.assertEqual(expected_username, storage_connection.rest_username)
        expected_password = mock_config.safe_get("dell_nas_password")
        self.assertEqual(expected_password, storage_connection.rest_password)
        expected_erify_certificates = mock_config.safe_get(
            "dell_ssl_cert_verify"
        )
        self.assertEqual(
            expected_erify_certificates, storage_connection.verify_certificate
        )

    def test_create_share_nfs(self):
        self._mock_powerflex_manager.get_storage_pool_id.return_value = (
            self.STORAGE_POOL_ID
        )
        self._mock_powerflex_manager.create_filesystem.return_value = (
            self.FILESYSTEM_ID
        )
        self._mock_powerflex_manager.create_nfs_export.return_value = (
            self.NFS_EXPORT_ID
        )
        self._mock_powerflex_manager.get_nfs_export_name.return_value = (
            self.NFS_EXPORT_NAME
        )
        self._mock_powerflex_manager.get_nas_server_interfaces.return_value = (
            [self.NAS_SERVER_IP]
        )

        self.assertFalse(
            self._mock_powerflex_manager.get_storage_pool_id.called
        )
        self.assertFalse(self._mock_powerflex_manager.create_filesystem.called)
        self.assertFalse(self._mock_powerflex_manager.create_nfs_export.called)
        self.assertFalse(
            self._mock_powerflex_manager.get_nfs_export_name.called
        )

        # create the share
        share = {"name": self.SHARE_NAME, "share_proto": "NFS", "size": 8}
        locations = self.storage_connection.create_share(
            self.mock_context, share, None
        )

        # verify location and API call made
        expected_locations = [{"path": "%s:/%s" % (
            self.NAS_SERVER_IP,
            self.SHARE_NAME,
        )}]
        self.assertEqual(expected_locations, locations)
        self._mock_powerflex_manager.get_storage_pool_id.assert_called_with(
            self._mock_config.safe_get("powerflex_protection_domain"),
            self._mock_config.safe_get("powerflex_storage_pool"),
        )
        self._mock_powerflex_manager.create_filesystem.assert_called_with(
            self.STORAGE_POOL_ID,
            self._mock_config.safe_get("dell_nas_server"),
            self.SHARE_NAME,
            8 * units.Gi,
        )
        self._mock_powerflex_manager.create_nfs_export.assert_called_with(
            self.FILESYSTEM_ID, self.SHARE_NAME
        )
        self._mock_powerflex_manager.get_nfs_export_name.assert_called_with(
            self.NFS_EXPORT_ID
        )

    def test_create_share_nfs_filesystem_id_not_found(self):
        share = {"name": self.SHARE_NAME, "share_proto": "NFS", "size": 8}
        self._mock_powerflex_manager.create_filesystem.return_value = None

        self.assertRaises(
            exception.ShareBackendException,
            self.storage_connection.create_share,
            self.mock_context,
            share,
            share_server=None,
        )

    def test_create_share_nfs_backend_failure(self):
        share = {"name": self.SHARE_NAME, "share_proto": "NFS", "size": 8}
        self._mock_powerflex_manager.create_nfs_export.return_value = False

        self.assertRaises(
            exception.ShareBackendException,
            self.storage_connection.create_share,
            self.mock_context,
            share,
            share_server=None,
        )

    def test_create_snapshot(self):
        self._mock_powerflex_manager.get_fsid_from_export_name.return_value = (
            self.FILESYSTEM_ID
        )
        self._mock_powerflex_manager.create_snapshot.return_value = True

        snapshot = {
            "name": self.SNAPSHOT_NAME,
            "share_name": self.SNAPSHOT_PATH,
            "id": self.SNAPSHOT_ID,
        }
        self.storage_connection.create_snapshot(
            self.mock_context, snapshot, None
        )

        # verify the create snapshot API call is executed
        self._mock_powerflex_manager.get_fsid_from_export_name. \
            assert_called_with(
                self.SNAPSHOT_PATH
            )
        self._mock_powerflex_manager.create_snapshot.assert_called_with(
            self.SNAPSHOT_NAME, self.FILESYSTEM_ID
        )

    def test_create_snapshot_failure(self):
        self._mock_powerflex_manager.get_fsid_from_export_name.return_value = (
            self.FILESYSTEM_ID
        )
        self._mock_powerflex_manager.create_snapshot.return_value = False

        snapshot = {
            "name": self.SNAPSHOT_NAME,
            "share_name": self.SNAPSHOT_PATH,
            "id": self.SNAPSHOT_ID,
        }
        self.storage_connection.create_snapshot(
            self.mock_context, snapshot, None
        )

    def test_delete_share_nfs(self):
        share = {"name": self.SHARE_NAME, "share_proto": "NFS"}

        self._mock_powerflex_manager.get_filesystem_id.return_value = (
            self.FILESYSTEM_ID
        )

        self.assertFalse(self._mock_powerflex_manager.get_filesystem_id.called)
        self.assertFalse(self._mock_powerflex_manager.delete_filesystem.called)

        # delete the share
        self.storage_connection.delete_share(self.mock_context, share, None)

        # verify share delete
        self._mock_powerflex_manager.get_filesystem_id.assert_called_with(
            self.SHARE_NAME
        )
        self._mock_powerflex_manager.delete_filesystem.assert_called_with(
            self.FILESYSTEM_ID
        )

    def test_delete_nfs_share_backend_failure(self):
        share = {"name": self.SHARE_NAME, "share_proto": "NFS"}

        self._mock_powerflex_manager.delete_filesystem.return_value = False
        self.assertRaises(
            exception.ShareBackendException,
            self.storage_connection.delete_share,
            self.mock_context,
            share,
            None,
        )

    def test_delete_nfs_share_share_does_not_exist(self):
        self._mock_powerflex_manager.get_filesystem_id.return_value = None
        share = {"name": self.SHARE_NAME, "share_proto": "NFS"}

        # verify the calling delete on a non-existent share returns and does
        # not throw exception
        self.storage_connection.delete_share(self.mock_context, share, None)
        self.assertTrue(self._mock_powerflex_manager.get_filesystem_id.called)
        self.assertFalse(self._mock_powerflex_manager.delete_filesystem.called)

    def test_delete_snapshot(self):
        self._mock_powerflex_manager.get_fsid_from_snapshot_name. \
            return_value = (
                self.FILESYSTEM_ID
            )

        self.assertFalse(
            self._mock_powerflex_manager.get_fsid_from_snapshot_name.called
        )
        self.assertFalse(self._mock_powerflex_manager.delete_filesystem.called)

        # delete the created snapshot
        snapshot = {
            "name": self.SNAPSHOT_NAME,
            "share_name": self.SNAPSHOT_PATH,
            "id": self.SNAPSHOT_ID,
        }
        self.storage_connection.delete_snapshot(
            self.mock_context, snapshot, None
        )

        # verify the API call was made to delete the snapshot
        self._mock_powerflex_manager.get_fsid_from_snapshot_name. \
            assert_called_with(
                self.SNAPSHOT_NAME
            )
        self._mock_powerflex_manager.delete_filesystem.assert_called_with(
            self.FILESYSTEM_ID
        )

    def test_delete_snapshot_backend_failure(self):
        self._mock_powerflex_manager.get_fsid_from_snapshot_name. \
            return_value = (
                self.FILESYSTEM_ID
            )
        self._mock_powerflex_manager.delete_filesystem.return_value = False

        self.assertFalse(
            self._mock_powerflex_manager.get_fsid_from_snapshot_name.called
        )
        self.assertFalse(self._mock_powerflex_manager.delete_filesystem.called)

        snapshot = {
            "name": self.SNAPSHOT_NAME,
            "share_name": self.SNAPSHOT_PATH,
            "id": self.SNAPSHOT_ID,
        }
        # verify the API call was made to delete the snapshot
        self.assertRaises(
            exception.ShareBackendException,
            self.storage_connection.delete_snapshot,
            self.mock_context,
            snapshot,
            None,
        )
        self._mock_powerflex_manager.get_fsid_from_snapshot_name. \
            assert_called_with(
                self.SNAPSHOT_NAME
            )
        self._mock_powerflex_manager.delete_filesystem.assert_called_with(
            self.FILESYSTEM_ID
        )

    def test_extend_share(self):
        new_share_size = 20
        share = {
            "name": self.SHARE_NAME,
            "share_proto": "NFS",
            "size": new_share_size,
        }
        self._mock_powerflex_manager.get_filesystem_id.return_value = (
            self.FILESYSTEM_ID
        )
        self.assertFalse(self._mock_powerflex_manager.get_filesystem_id.called)

        self.storage_connection.extend_share(share, new_share_size)

        self._mock_powerflex_manager.get_filesystem_id.assert_called_with(
            self.SHARE_NAME
        )
        expected_quota_size = new_share_size * units.Gi
        self._mock_powerflex_manager.extend_export.assert_called_once_with(
            self.FILESYSTEM_ID, expected_quota_size
        )

    def test_update_access_add_nfs(self):
        share = {"name": self.SHARE_NAME, "share_proto": "NFS"}

        self._mock_powerflex_manager.get_nfs_export_id.return_value = (
            self.NFS_EXPORT_ID
        )
        self._mock_powerflex_manager.set_export_access.return_value = True

        self.assertFalse(self._mock_powerflex_manager.get_nfs_export_id.called)
        self.assertFalse(self._mock_powerflex_manager.set_export_access.called)

        nfs_rw_ip = "192.168.0.10"
        nfs_ro_ip = "192.168.0.11"
        nfs_access_rw = {
            "access_type": "ip",
            "access_to": nfs_rw_ip,
            "access_level": const.ACCESS_LEVEL_RW,
            "access_id": "09960614-8574-4e03-89cf-7cf267b0bd08",
        }
        nfs_access_ro = {
            "access_type": "ip",
            "access_to": nfs_ro_ip,
            "access_level": const.ACCESS_LEVEL_RO,
            "access_id": "09960614-8574-4e03-89cf-7cf267b0bd08",
        }
        access_rules = [nfs_access_rw, nfs_access_ro]

        self.storage_connection.update_access(
            self.mock_context,
            share,
            access_rules,
            add_rules=None,
            delete_rules=None,
            share_server=None,
        )

        self._mock_powerflex_manager.get_nfs_export_id.assert_called_once_with(
            self.SHARE_NAME
        )
        self._mock_powerflex_manager.set_export_access.assert_called_once_with(
            self.NFS_EXPORT_ID, {nfs_rw_ip}, {nfs_ro_ip}
        )

    def test_update_access_add_nfs_invalid_acess_type(self):
        share = {
            "name": self.SHARE_NAME,
            "share_proto": "NFS",
            "display_name": "foo_display_name",
        }

        nfs_rw_ip = "192.168.0.10"
        nfs_ro_ip = "192.168.0.11"
        nfs_access_rw = {
            "access_type": "invalid_type",
            "access_to": nfs_rw_ip,
            "access_level": const.ACCESS_LEVEL_RW,
            "access_id": "09960614-8574-4e03-89cf-7cf267b0bd08",
        }
        nfs_access_ro = {
            "access_type": "invalid_type",
            "access_to": nfs_ro_ip,
            "access_level": const.ACCESS_LEVEL_RO,
            "access_id": "09960614-8574-4e03-89cf-7cf267b0bd09",
        }
        access_rules = [nfs_access_rw, nfs_access_ro]

        self._mock_powerflex_manager.get_nfs_export_id.return_value = (
            self.NFS_EXPORT_ID
        )

        access_updates = self.storage_connection.update_access(
            self.mock_context,
            share,
            access_rules,
            add_rules=None,
            delete_rules=None,
            share_server=None,
        )

        self._mock_powerflex_manager.set_export_access.assert_called_once_with(
            self.NFS_EXPORT_ID, set(), set()
        )

        self.assertIsNotNone(access_updates)

    def test_update_access_add_nfs_backend_failure(self):
        share = {
            "name": self.SHARE_NAME,
            "share_proto": "NFS",
            "display_name": "foo_display_name",
        }

        self._mock_powerflex_manager.get_nfs_export_id.return_value = (
            self.NFS_EXPORT_ID
        )
        self._mock_powerflex_manager.set_export_access.return_value = False

        self.assertFalse(self._mock_powerflex_manager.get_nfs_export_id.called)
        self.assertFalse(self._mock_powerflex_manager.set_export_access.called)

        nfs_rw_ip = "192.168.0.10"
        nfs_ro_ip = "192.168.0.11"
        nfs_access_rw = {
            "access_type": "ip",
            "access_to": nfs_rw_ip,
            "access_level": const.ACCESS_LEVEL_RW,
            "access_id": "09960614-8574-4e03-89cf-7cf267b0bd08",
        }
        nfs_access_ro = {
            "access_type": "ip",
            "access_to": nfs_ro_ip,
            "access_level": const.ACCESS_LEVEL_RO,
            "access_id": "09960614-8574-4e03-89cf-7cf267b0bd08",
        }
        access_rules = [nfs_access_rw, nfs_access_ro]

        self.assertRaises(
            exception.ShareBackendException,
            self.storage_connection.update_access,
            self.mock_context,
            share,
            access_rules,
            add_rules=None,
            delete_rules=None,
            share_server=None,
        )

    def test_update_share_stats(self):
        data = dict(
            share_backend_name='powerflex',
            vendor_name='Dell EMC',
            storage_protocol='NFS_CIFS',
            snapshot_support=True,
            create_share_from_snapshot_support=True)
        stats = dict(
            maxCapacityInKb=4826330112,
            capacityInUseInKb=53217280,
            netUnusedCapacityInKb=1566080512,
            primaryVacInKb=184549376)

        self._mock_powerflex_manager.get_storage_pool_id.return_value = (
            self.STORAGE_POOL_ID
        )
        self._mock_powerflex_manager.get_storage_pool_statistic. \
            return_value = stats
        self.storage_connection.update_share_stats(data)
        self.assertEqual(data['storage_protocol'], 'NFS')
        self.assertEqual(data['create_share_from_snapshot_support'], False)
        self.assertEqual(data['driver_version'], connection.VERSION)
        self.assertIsNotNone(data['pools'])

    def test_get_default_filter_function(self):
        filter = self.storage_connection.get_default_filter_function()
        self.assertEqual(filter, "share.size >= 3")

    def test_create_share_from_snapshot(self):
        self.assertRaises(
            NotImplementedError,
            self.storage_connection.create_share_from_snapshot,
            self.mock_context,
            share=None,
            snapshot=None,
        )

    def test_allow_access(self):
        self.assertRaises(
            NotImplementedError,
            self.storage_connection.allow_access,
            self.mock_context,
            share=None,
            access=None,
            share_server=None,
        )

    def test_deny_access(self):
        self.assertRaises(
            NotImplementedError,
            self.storage_connection.deny_access,
            self.mock_context,
            share=None,
            access=None,
            share_server=None,
        )

    def test_setup_server(self):
        self.assertRaises(
            NotImplementedError,
            self.storage_connection.setup_server,
            network_info=None,
        )

    def test_teardown_server(self):
        self.assertRaises(
            NotImplementedError,
            self.storage_connection.teardown_server,
            server_details=None,
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""API request/response validating middleware."""

import functools
import typing as ty

from oslo_serialization import jsonutils
import webob

from manila.api.openstack import api_version_request as api_version
from manila.api.openstack import wsgi
from manila.api.validation import validators
from manila import exception
from manila.i18n import _


def validated(cls):
    cls._validated = True
    return cls


def _schema_validator(
    schema: ty.Dict[str, ty.Any],
    target: ty.Dict[str, ty.Any],
    min_version: ty.Optional[str],
    max_version: ty.Optional[str],
    args: ty.Any,
    kwargs: ty.Any,
    is_body: bool = True,
):
    """A helper method to execute JSON Schema Validation.

    This method checks the request version whether matches the specified
    ``max_version`` and ``min_version``. If the version range matches the
    request, we validate ``schema`` against ``target``. A failure will result
    in ``ValidationError`` being raised.

    :param schema: The JSON Schema schema used to validate the target.
    :param target: The target to be validated by the schema.
    :param min_version: A string indicating the minimum API version ``schema``
        applies against.
    :param max_version: A string indicating the maximum API version ``schema``
        applies against.
    :param args: Positional arguments which passed into original method.
    :param kwargs: Keyword arguments which passed into original method.
    :param is_body: Whether ``target`` is a HTTP request body or not.
    :returns: None.
    :raises: ``ValidationError`` if validation fails.
    """
    min_ver = api_version.APIVersionRequest(min_version)
    max_ver = api_version.APIVersionRequest(max_version)

    # NOTE: The request object is always the second argument. However, numerous
    # unittests pass in the request object via kwargs instead so we handle that
    # as well.
    # TODO(stephenfin): Fix unit tests so we don't have to to do this
    if 'req' in kwargs:
        ver = kwargs['req'].api_version_request
    else:
        ver = args[1].api_version_request

    if ver.matches(min_ver, max_ver):
        # Only validate against the schema if it lies within
        # the version range specified. Note that if both min
        # and max are not specified the validator will always
        # be run.
        schema_validator = validators._SchemaValidator(schema, is_body=is_body)
        schema_validator.validate(target)


def request_body_schema(
    schema: ty.Dict[str, ty.Any],
    min_version: ty.Optional[str] = None,
    max_version: ty.Optional[str] = None,
):
    """Register a schema to validate request body.

    ``schema`` will be used for validating the request body just before the API
    method is executed.

    :param schema: The JSON Schema schema used to validate the target.
    :param min_version: A string indicating the minimum API version ``schema``
        applies against.
    :param max_version: A string indicating the maximum API version ``schema``
        applies against.
    """

    def add_validator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            _schema_validator(
                schema,
                kwargs['body'],
                min_version,
                max_version,
                args,
                kwargs,
                is_body=True,
            )
            return func(*args, **kwargs)

        wrapper._request_body_schema = schema

        return wrapper

    return add_validator


def request_query_schema(
    schema: ty.Dict[str, ty.Any],
    min_version: ty.Optional[str] = None,
    max_version: ty.Optional[str] = None,
):
    """Register a schema to validate request query string parameters.

    ``schema`` will be used for validating request query strings just before
    the API method is executed.

    :param schema: The JSON Schema schema used to validate the target.
    :param min_version: A string indicating the minimum API version ``schema``
        applies against.
    :param max_version: A string indicating the maximum API version ``schema``
        applies against.
    """

    def add_validator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # NOTE: The request object is always the second argument. However,
            # numerous unittests pass in the request object via kwargs instead
            # so we handle that as well.
            # TODO(stephenfin): Fix unit tests so we don't have to to do this
            if 'req' in kwargs:
                req = kwargs['req']
            else:
                req = args[1]

            # NOTE: The webob package throws UnicodeError when param cannot be
            # decoded. Catch this and raise HTTP 400.
            try:
                query = req.GET.dict_of_lists()
            except UnicodeDecodeError:
                msg = _('Query string is not UTF-8 encoded')
                raise exception.ValidationError(msg)

            _schema_validator(
                schema,
                query,
                min_version,
                max_version,
                args,
                kwargs,
                is_body=True,
            )
            return func(*args, **kwargs)

        wrapper._request_query_schema = schema

        return wrapper

    return add_validator


def response_body_schema(
    schema: ty.Dict[str, ty.Any],
    min_version: ty.Optional[str] = None,
    max_version: ty.Optional[str] = None,
):
    """Register a schema to validate response body.

    ``schema`` will be used for validating the response body just after the API
    method is executed.

    :param schema: The JSON Schema schema used to validate the target.
    :param min_version: A string indicating the minimum API version ``schema``
        applies against.
    :param max_version: A string indicating the maximum API version ``schema``
        applies against.
    """

    def add_validator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            response = func(*args, **kwargs)

            # NOTE(stephenfin): If our response is an object, we need to
            # serializer and deserialize to convert e.g. date-time to strings
            if isinstance(response, wsgi.ResponseObject):
                serializer = wsgi.JSONDictSerializer()
                _body = serializer.serialize(response.obj)
            # TODO(stephenfin): We should replace all instances of this with
            # wsgi.ResponseObject
            elif isinstance(response, webob.Response):
                _body = response.body
            else:
                serializer = wsgi.JSONDictSerializer()
                _body = serializer.serialize(response)

            if _body == b'':
                body = None
            else:
                body = jsonutils.loads(_body)

            _schema_validator(
                schema,
                body,
                min_version,
                max_version,
                args,
                kwargs,
                is_body=True,
            )
            return response

        wrapper._response_body_schema = schema

        return wrapper

    return add_validator



# Copyright 2024 VAST Data Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import ipaddress
import types

from oslo_config import cfg
from oslo_log import log
from oslo_utils import timeutils


CONF = cfg.CONF
LOG = log.getLogger(__name__)


class Bunch(dict):
    # from https://github.com/real-easypy/easypy

    __slots__ = ("__stop_recursing__",)

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            if name[0] == "_" and name[1:].isdigit():
                return self[name[1:]]
            raise AttributeError(
                "%s has no attribute %r" % (self.__class__, name)
            )

    def __getitem__(self, key):
        try:
            return super(Bunch, self).__getitem__(key)
        except KeyError:
            from numbers import Integral

            if isinstance(key, Integral):
                return self[str(key)]
            raise

    def __setattr__(self, name, value):
        self[name] = value

    def __delattr__(self, name):
        try:
            del self[name]
        except KeyError:
            raise AttributeError(
                "%s has no attribute %r" % (self.__class__, name)
            )

    def __getstate__(self):
        return self

    def __setstate__(self, dict):
        self.update(dict)

    def __repr__(self):
        if getattr(self, "__stop_recursing__", False):
            items = sorted(
                "%s" % k for k in self
                if isinstance(k, str) and not k.startswith("__")
            )
            attrs = ", ".join(items)
        else:
            dict.__setattr__(self, "__stop_recursing__", True)
            try:
                attrs = self.render()
            finally:
                dict.__delattr__(self, "__stop_recursing__")
        return "%s(%s)" % (self.__class__.__name__, attrs)

    def render(self):
        items = sorted(
            "%s=%r" % (k, v)
            for k, v in self.items()
            if isinstance(k, str) and not k.startswith("__")
        )
        return ", ".join(items)

    def to_dict(self):
        return unbunchify(self)

    def to_json(self):
        import json

        return json.dumps(self.to_dict())

    def copy(self, deep=False):
        if deep:
            return _convert(self, self.__class__)
        else:
            return self.__class__(self)

    @classmethod
    def from_dict(cls, d):
        return _convert(d, cls)

    @classmethod
    def from_json(cls, d):
        import json

        return cls.from_dict(json.loads(d))

    def __dir__(self):
        members = set(
            k
            for k in self
            if isinstance(k, str)
            and (k[0] == "_" or k.replace("_", "").isalnum())
        )
        members.update(dict.__dir__(self))
        return sorted(members)

    def without(self, *keys):
        "Return a shallow copy of the bunch without the specified keys"
        return Bunch((k, v) for k, v in self.items() if k not in keys)

    def but_with(self, **kw):
        "Return a shallow copy of the bunch with the specified keys"
        return Bunch(self, **kw)


def _convert(d, typ):
    if isinstance(d, dict):
        return typ({str(k): _convert(v, typ) for k, v in d.items()})
    elif isinstance(d, (tuple, list, set)):
        return type(d)(_convert(e, typ) for e in d)
    else:
        return d


def unbunchify(d):
    """Recursively convert Bunches in `d` to a regular dicts."""
    return _convert(d, dict)


def bunchify(d=None, **kw):
    """Recursively convert dicts in `d` to Bunches.

    If `kw` given, recursively convert dicts in
    it to Bunches and update `d` with it.
    If `d` is None, an empty Bunch is made.
    """

    d = _convert(d, Bunch) if d is not None else Bunch()
    if kw:
        d.update(bunchify(kw))
    return d


def generate_ip_range(ip_ranges):
    """Generate list of ips from provided ip ranges.

    `ip_ranges` should be list of ranges where fist
    ip in range represents start ip and second is end ip
    eg: [["15.0.0.1", "15.0.0.4"], ["10.0.0.27", "10.0.0.30"]]
    """
    return [
        ip.compressed
        for start_ip, end_ip in ip_ranges
        for net in ipaddress.summarize_address_range(
            ipaddress.ip_address(start_ip),
            ipaddress.ip_address(end_ip)
        )
        for ip in net
    ]


def decorate_methods_with(dec):
    if not CONF.debug:
        return lambda cls: cls

    def inner(cls):
        for attr_name, attr_val in cls.__dict__.items():
            if (isinstance(attr_val, types.FunctionType) and
                    not attr_name.startswith("_")):
                setattr(cls, attr_name, dec(attr_val))
        return cls

    return inner


def verbose_driver_trace(fn):
    if not CONF.debug:
        return fn

    def inner(self, *args, **kwargs):
        start = timeutils.utcnow()
        LOG.debug(f"[{fn.__name__}] >>>")
        res = fn(self, *args, **kwargs)
        end = timeutils.utcnow()
        LOG.debug(
            f"Spent {timeutils.delta_seconds(start, end)} sec. "
            f"Return {res}.\n"
            f"<<< [{fn.__name__}]"
        )
        return res

    return inner



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

index_request_query = {
    'type': 'object',
    'properties': {},
    # TODO(stephenfin): Exclude additional query string parameters in a future
    # microversion
    'additionalProperties': True,
}

index_response_body = {
    'type': 'object',
    'properties': {
        'availability_zones': {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    'id': {
                        'type': 'string',
                        'format': 'uuid',
                    },
                    'name': {
                        'type': 'string',
                        'minLength': 1,
                        'maxLength': 255,
                    },
                    'created_at': {
                        'type': 'string',
                        'format': 'date-time',
                    },
                    'updated_at': {
                        'type': ['string', 'null'],
                        'format': 'date-time',
                    },
                },
                'required': [],
                'additionalProperties': False,
            },
        },
    },
    'required': ['availability_zones'],
    'additionalProperties': False,
}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from manila.api import common


class ViewBuilder(common.ViewBuilder):
    """Model a resource lock API response as a python dictionary."""

    _collection_name = "resource_locks"

    def index(self, request, resource_locks, count=None):
        """Show a list of resource locks."""
        return self._list_view(self.detail,
                               request,
                               resource_locks,
                               count=count)

    def detail(self, request, resource_lock):
        """Detailed view of a single resource lock."""
        lock_ref = {
            'id': resource_lock.get('id'),
            'user_id': resource_lock.get('user_id'),
            'project_id': resource_lock.get('project_id'),
            'lock_context': resource_lock.get('lock_context'),
            'resource_type': resource_lock.get('resource_type'),
            'resource_id': resource_lock.get('resource_id'),
            'resource_action': resource_lock.get('resource_action'),
            'lock_reason': resource_lock.get('lock_reason'),
            'created_at': resource_lock.get('created_at'),
            'updated_at': resource_lock.get('updated_at'),
            'links': self._get_links(request, resource_lock['id']),
        }
        return {'resource_lock': lock_ref}

    def _list_view(self, func, request, resource_locks,
                   coll_name=_collection_name, count=None):
        """Provide a view for a list of resource_locks.

        :param func: Function used to format the lock data
        :param request: API request
        :param resource_locks: List of locks in dictionary format
        :param coll_name: Name of collection, used to generate the next link
                          for a pagination query
        :returns: lock data in dictionary format
        """
        locks_list = [
            func(request, lock)['resource_lock']
            for lock in resource_locks
        ]
        locks_links = self._get_collection_links(request,
                                                 resource_locks,
                                                 coll_name)
        locks_dict = dict({"resource_locks": locks_list})

        if count:
            locks_dict['count'] = count

        if locks_links:
            locks_dict['resource_locks_links'] = locks_links

        return locks_dict



# Copyright (c) 2023 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from unittest import mock

from oslo_log import log
from oslo_utils import units

from manila.common import constants as const
from manila import exception
from manila.share.drivers.dell_emc.plugins.powerstore import connection
from manila import test

LOG = log.getLogger(__name__)


class PowerStoreTest(test.TestCase):
    """Unit test for the PowerStore Manila driver."""

    REST_IP = "192.168.0.110"
    NAS_SERVER_NAME = "powerstore-nasserver"
    NAS_SERVER_ID = "6423d56e-eaf3-7424-be0b-1a9efb93188b"
    NAS_SERVER_IP = "192.168.11.23"
    SHARE_NAME = "powerstore-share"
    SHARE_SIZE_GB = 3
    SHARE_NEW_SIZE_GB = 6
    FILESYSTEM_ID = "6454e9a9-a698-e9bc-ca61-1a9efb93188b"
    NFS_EXPORT_ID = "6454ec18-7b8d-1532-1b8a-1a9efb93188b"
    SMB_SHARE_ID = "64927ae9-3403-6930-a784-f227b9987c54"
    RW_HOSTS = "192.168.1.10"
    RO_HOSTS = "192.168.1.11"
    RW_USERS = "user_1"
    RO_USERS = "user_2"
    SNAPSHOT_NAME = "powerstore-share-snap"
    SNAPSHOT_ID = "6454ea29-09c3-030e-cfc3-1a9efb93188b"
    CLONE_ID = "64560f05-e677-ec2a-7fcf-1a9efb93188b"
    CLONE_NAME = "powerstore-nfs-share-snap-clone"

    class MockConfig(object):
        def safe_get(self, value):
            if value == "dell_nas_backend_host":
                return "192.168.0.110"
            elif value == "dell_nas_login":
                return "admin"
            elif value == "dell_nas_password":
                return "pwd"
            elif value == "dell_nas_server":
                return "powerstore-nasserver"
            elif value == "dell_ad_domain":
                return "domain_name"
            elif value == "dell_ssl_cert_verify":
                return True
            elif value == "dell_ssl_cert_path":
                return "powerstore_cert_path"

    @mock.patch(
        "manila.share.drivers.dell_emc.plugins.powerstore.client."
        "PowerStoreClient",
        autospec=True,
    )
    def setUp(self, mock_powerstore_client):
        super(PowerStoreTest, self).setUp()

        self._mock_powerstore_client = mock_powerstore_client.return_value
        self.storage_connection = connection.PowerStoreStorageConnection(LOG)

        self.mock_context = mock.Mock("Context")
        self.mock_emc_driver = mock.Mock("EmcDriver")

        self._mock_config = self.MockConfig()
        self.mock_emc_driver.attach_mock(self._mock_config, "configuration")
        self.storage_connection.connect(
            self.mock_emc_driver, self.mock_context
        )

    def test_connect(self):
        storage_connection = connection.PowerStoreStorageConnection(LOG)

        # execute method under test
        storage_connection.connect(self.mock_emc_driver, self.mock_context)

        # verify connect sets driver params appropriately
        mock_config = self.MockConfig()
        server_addr = mock_config.safe_get("dell_nas_backend_host")
        self.assertEqual(server_addr, storage_connection.rest_ip)
        expected_username = mock_config.safe_get("dell_nas_login")
        self.assertEqual(expected_username, storage_connection.rest_username)
        expected_password = mock_config.safe_get("dell_nas_password")
        self.assertEqual(expected_password, storage_connection.rest_password)
        expected_nas_server = mock_config.safe_get("dell_nas_server")
        self.assertEqual(expected_nas_server, storage_connection.nas_server)
        expected_ad_domain = mock_config.safe_get("dell_ad_domain")
        self.assertEqual(expected_ad_domain, storage_connection.ad_domain)
        expected_verify_certificate = mock_config.safe_get(
            "dell_ssl_cert_verify"
        )
        self.assertEqual(
            expected_verify_certificate, storage_connection.verify_certificate
        )

    def test_create_share_nfs(self):
        self._mock_powerstore_client.get_nas_server_id.return_value = (
            self.NAS_SERVER_ID
        )
        self._mock_powerstore_client.create_filesystem.return_value = (
            self.FILESYSTEM_ID
        )
        self._mock_powerstore_client.create_nfs_export.return_value = (
            self.NFS_EXPORT_ID
        )
        self._mock_powerstore_client.get_nas_server_interfaces.return_value = (
            [{"ip": self.NAS_SERVER_IP, "preferred": True}]
        )

        self.assertFalse(self._mock_powerstore_client.get_nas_server_id.called)
        self.assertFalse(self._mock_powerstore_client.create_filesystem.called)
        self.assertFalse(self._mock_powerstore_client.create_nfs_export.called)
        self.assertFalse(
            self._mock_powerstore_client.get_nas_server_interfaces.called
        )

        # create the share
        share = {"name": self.SHARE_NAME, "share_proto": "NFS",
                 "size": self.SHARE_SIZE_GB}
        locations = self.storage_connection.create_share(
            self.mock_context,
            share,
            None
        )

        # verify location and API call made
        expected_locations = [
            {"path": "%s:/%s" % (
                self.NAS_SERVER_IP,
                self.SHARE_NAME,
                ),
             "metadata": {
                 "preferred": True}}]
        self.assertEqual(expected_locations, locations)
        self._mock_powerstore_client.get_nas_server_id.assert_called_with(
            self._mock_config.safe_get("dell_nas_server")
        )
        self._mock_powerstore_client.create_filesystem.assert_called_with(
            self.NAS_SERVER_ID,
            self.SHARE_NAME,
            self.SHARE_SIZE_GB * units.Gi,
        )
        self._mock_powerstore_client.create_nfs_export.assert_called_with(
            self.FILESYSTEM_ID, self.SHARE_NAME
        )
        self._mock_powerstore_client.get_nas_server_interfaces. \
            assert_called_with(
                self.NAS_SERVER_ID
            )

    def test_create_share_cifs(self):
        self._mock_powerstore_client.get_nas_server_id.return_value = (
            self.NAS_SERVER_ID
        )
        self._mock_powerstore_client.create_filesystem.return_value = (
            self.FILESYSTEM_ID
        )
        self._mock_powerstore_client.create_smb_share.return_value = (
            self.SMB_SHARE_ID
        )
        self._mock_powerstore_client.get_nas_server_interfaces.return_value = (
            [{"ip": self.NAS_SERVER_IP, "preferred": True}]
        )

        self.assertFalse(self._mock_powerstore_client.get_nas_server_id.called)
        self.assertFalse(self._mock_powerstore_client.create_filesystem.called)
        self.assertFalse(self._mock_powerstore_client.create_smb_share.called)
        self.assertFalse(
            self._mock_powerstore_client.get_nas_server_interfaces.called
        )

        # create the share
        share = {"name": self.SHARE_NAME, "share_proto": "CIFS",
                 "size": self.SHARE_SIZE_GB}
        locations = self.storage_connection.create_share(
            self.mock_context,
            share,
            None
        )

        # verify location and API call made
        expected_locations = [
            {"path": "\\\\%s\\%s" % (
                self.NAS_SERVER_IP,
                self.SHARE_NAME),
             "metadata": {
                 "preferred": True}}]
        self.assertEqual(expected_locations, locations)
        self._mock_powerstore_client.get_nas_server_id.assert_called_with(
            self._mock_config.safe_get("dell_nas_server")
        )
        self._mock_powerstore_client.create_filesystem.assert_called_with(
            self.NAS_SERVER_ID,
            self.SHARE_NAME,
            self.SHARE_SIZE_GB * units.Gi,
        )
        self._mock_powerstore_client.create_smb_share.assert_called_with(
            self.FILESYSTEM_ID, self.SHARE_NAME
        )
        self._mock_powerstore_client.get_nas_server_interfaces. \
            assert_called_with(
                self.NAS_SERVER_ID
            )

    def test_create_share_filesystem_id_not_found(self):
        share = {"name": self.SHARE_NAME, "share_proto": "NFS",
                 "size": self.SHARE_SIZE_GB}
        self._mock_powerstore_client.create_filesystem.return_value = None

        self.assertRaises(
            exception.ShareBackendException,
            self.storage_connection.create_share,
            self.mock_context,
            share,
            share_server=None
        )

    def test_create_share_nfs_backend_failure(self):
        share = {"name": self.SHARE_NAME, "share_proto": "NFS",
                 "size": self.SHARE_SIZE_GB}
        self._mock_powerstore_client.create_nfs_export.return_value = None

        self.assertRaises(
            exception.ShareBackendException,
            self.storage_connection.create_share,
            self.mock_context,
            share,
            share_server=None
        )

    def test_create_share_cifs_backend_failure(self):
        share = {"name": self.SHARE_NAME, "share_proto": "CIFS",
                 "size": self.SHARE_SIZE_GB}
        self._mock_powerstore_client.create_smb_share.return_value = None

        self.assertRaises(
            exception.ShareBackendException,
            self.storage_connection.create_share,
            self.mock_context,
            share,
            share_server=None
        )

    def test_delete_share_nfs(self):
        share = {"name": self.SHARE_NAME, "share_proto": "NFS"}

        self._mock_powerstore_client.get_filesystem_id.return_value = (
            self.FILESYSTEM_ID
        )

        self.assertFalse(self._mock_powerstore_client.get_filesystem_id.called)
        self.assertFalse(self._mock_powerstore_client.delete_filesystem.called)

        # delete the share
        self.storage_connection.delete_share(self.mock_context, share, None)

        # verify share delete
        self._mock_powerstore_client.get_filesystem_id.assert_called_with(
            self.SHARE_NAME
        )
        self._mock_powerstore_client.delete_filesystem.assert_called_with(
            self.FILESYSTEM_ID
        )

    def test_delete_nfs_share_backend_failure(self):
        share = {"name": self.SHARE_NAME, "share_proto": "NFS"}

        self._mock_powerstore_client.delete_filesystem.return_value = False
        self.assertRaises(
            exception.ShareBackendException,
            self.storage_connection.delete_share,
            self.mock_context,
            share,
            None,
        )

    def test_delete_nfs_share_share_does_not_exist(self):
        self._mock_powerstore_client.get_filesystem_id.return_value = None
        share = {"name": self.SHARE_NAME, "share_proto": "NFS"}

        # verify the calling delete on a non-existent share returns and does
        # not throw exception
        self.storage_connection.delete_share(self.mock_context, share, None)
        self.assertTrue(self._mock_powerstore_client.get_filesystem_id.called)
        self.assertFalse(self._mock_powerstore_client.delete_filesystem.called)

    def test_extend_share(self):
        share = {
            "name": self.SHARE_NAME,
            "share_proto": "NFS",
            "size": self.SHARE_NEW_SIZE_GB,
        }
        self._mock_powerstore_client.get_filesystem_id.return_value = (
            self.FILESYSTEM_ID
        )
        self._mock_powerstore_client.resize_filesystem.return_value = (
            True, None
        )
        self.assertFalse(self._mock_powerstore_client.get_filesystem_id.called)

        self.storage_connection.extend_share(share, self.SHARE_NEW_SIZE_GB,
                                             self.NAS_SERVER_NAME)

        self._mock_powerstore_client.get_filesystem_id.assert_called_with(
            self.SHARE_NAME
        )
        expected_quota_size = self.SHARE_NEW_SIZE_GB * units.Gi
        self._mock_powerstore_client.resize_filesystem.assert_called_once_with(
            self.FILESYSTEM_ID, expected_quota_size
        )

    def test_shrink_share(self):
        share = {
            "name": self.SHARE_NAME,
            "share_proto": "NFS",
            "size": self.SHARE_SIZE_GB,
        }
        self._mock_powerstore_client.get_filesystem_id.return_value = (
            self.FILESYSTEM_ID
        )
        self._mock_powerstore_client.resize_filesystem.return_value = (
            True, None
        )
        self.assertFalse(self._mock_powerstore_client.get_filesystem_id.called)

        self.storage_connection.shrink_share(share, self.SHARE_NEW_SIZE_GB,
                                             self.NAS_SERVER_NAME)

        self._mock_powerstore_client.get_filesystem_id.assert_called_with(
            self.SHARE_NAME
        )
        expected_quota_size = self.SHARE_NEW_SIZE_GB * units.Gi
        self._mock_powerstore_client.resize_filesystem.assert_called_once_with(
            self.FILESYSTEM_ID, expected_quota_size
        )

    def test_shrink_share_failure(self):
        share = {
            "name": self.SHARE_NAME,
            "share_proto": "NFS",
            "size": self.SHARE_SIZE_GB,
            "id": self.CLONE_ID
        }
        self._mock_powerstore_client.get_filesystem_id.return_value = (
            self.FILESYSTEM_ID
        )
        self._mock_powerstore_client.resize_filesystem.return_value = (
            False, "msg"
        )

        self.assertRaises(
            exception.ShareShrinkingPossibleDataLoss,
            self.storage_connection.shrink_share,
            share,
            self.SHARE_NEW_SIZE_GB,
            self.NAS_SERVER_NAME
        )

    def test_shrink_share_backend_failure(self):
        share = {
            "name": self.SHARE_NAME,
            "share_proto": "NFS",
            "size": self.SHARE_SIZE_GB,
        }
        self._mock_powerstore_client.get_filesystem_id.return_value = (
            self.FILESYSTEM_ID
        )
        self._mock_powerstore_client.resize_filesystem.return_value = (
            False, None
        )

        self.assertRaises(
            exception.ShareBackendException,
            self.storage_connection.shrink_share,
            share,
            self.SHARE_NEW_SIZE_GB,
            self.NAS_SERVER_NAME
        )

    def test_update_access_add_nfs(self):
        share = {"name": self.SHARE_NAME, "share_proto": "NFS"}

        self._mock_powerstore_client.get_nfs_export_id.return_value = (
            self.NFS_EXPORT_ID
        )
        self._mock_powerstore_client.set_export_access.return_value = True

        self.assertFalse(self._mock_powerstore_client.get_nfs_export_id.called)
        self.assertFalse(self._mock_powerstore_client.set_export_access.called)

        nfs_access_rw = {
            "access_type": "ip",
            "access_to": self.RW_HOSTS,
            "access_level": const.ACCESS_LEVEL_RW,
            "access_id": "09960614-8574-4e03-89cf-7cf267b0bd08",
        }
        nfs_access_ro = {
            "access_type": "ip",
            "access_to": self.RO_HOSTS,
            "access_level": const.ACCESS_LEVEL_RO,
            "access_id": "09960614-8574-4e03-89cf-7cf267b0bd08",
        }
        access_rules = [nfs_access_rw, nfs_access_ro]

        self.storage_connection.update_access(
            self.mock_context,
            share,
            access_rules,
            add_rules=None,
            delete_rules=None,
            share_server=None,
        )

        self._mock_powerstore_client.get_nfs_export_id.assert_called_once_with(
            self.SHARE_NAME
        )
        self._mock_powerstore_client.set_export_access.assert_called_once_with(
            self.NFS_EXPORT_ID, {self.RW_HOSTS}, {self.RO_HOSTS}
        )

    def test_update_access_add_cifs(self):
        share = {"name": self.SHARE_NAME, "share_proto": "CIFS"}

        self._mock_powerstore_client.get_smb_share_id.return_value = (
            self.SMB_SHARE_ID
        )
        self._mock_powerstore_client.set_acl.return_value = True

        self.assertFalse(self._mock_powerstore_client.get_smb_share_id.called)
        self.assertFalse(self._mock_powerstore_client.set_acl.called)

        cifs_access_rw = {
            "access_type": "user",
            "access_to": self.RW_USERS,
            "access_level": const.ACCESS_LEVEL_RW,
            "access_id": "09960614-8574-4e03-89cf-7cf267b0bd08",
        }
        cifs_access_ro = {
            "access_type": "user",
            "access_to": self.RO_USERS,
            "access_level": const.ACCESS_LEVEL_RO,
            "access_id": "09960614-8574-4e03-89cf-7cf267b0bd08",
        }
        access_rules = [cifs_access_rw, cifs_access_ro]

        self.storage_connection.update_access(
            self.mock_context,
            share,
            access_rules,
            add_rules=None,
            delete_rules=None,
            share_server=None,
        )

        self._mock_powerstore_client.get_smb_share_id.assert_called_once_with(
            self.SHARE_NAME
        )
        self._mock_powerstore_client.set_acl.assert_called_once_with(
            self.SMB_SHARE_ID, {'domain_name\\user_1'}, {'domain_name\\user_2'}
        )

    def test_update_access_invalid_prefix(self):
        share = {"name": self.SHARE_NAME, "share_proto": "CIFS"}

        self._mock_powerstore_client.get_smb_share_id.return_value = (
            self.SMB_SHARE_ID
        )
        self._mock_powerstore_client.get_nas_server_smb_netbios. \
            return_value = None

        self.assertFalse(self._mock_powerstore_client.get_smb_share_id.called)
        self.assertFalse(self._mock_powerstore_client.set_acl.called)

        cifs_access_rw = {
            "access_type": "user",
            "access_to": self.RW_USERS,
            "access_level": const.ACCESS_LEVEL_RW,
            "access_id": "09960614-8574-4e03-89cf-7cf267b0bd08",
        }
        cifs_access_ro = {
            "access_type": "user",
            "access_to": self.RO_USERS,
            "access_level": const.ACCESS_LEVEL_RO,
            "access_id": "09960614-8574-4e03-89cf-7cf267b0bd08",
        }
        access_rules = [cifs_access_rw, cifs_access_ro]

        self.storage_connection.ad_domain = None

        access_updates = self.storage_connection.update_access(
            self.mock_context,
            share,
            access_rules,
            add_rules=None,
            delete_rules=None,
            share_server=None,
        )

        self._mock_powerstore_client.set_acl.assert_called_once_with(
            self.SMB_SHARE_ID, set(), set()
        )

        self.assertIsNotNone(access_updates)

    def test_update_access_add_nfs_invalid_acess_type(self):
        share = {
            "name": self.SHARE_NAME,
            "share_proto": "NFS",
            "display_name": "foo_display_name",
        }

        nfs_access_rw = {
            "access_type": "invalid_type",
            "access_to": self.RW_HOSTS,
            "access_level": const.ACCESS_LEVEL_RW,
            "access_id": "09960614-8574-4e03-89cf-7cf267b0bd08",
        }
        nfs_access_ro = {
            "access_type": "invalid_type",
            "access_to": self.RO_HOSTS,
            "access_level": const.ACCESS_LEVEL_RO,
            "access_id": "09960614-8574-4e03-89cf-7cf267b0bd09",
        }
        access_rules = [nfs_access_rw, nfs_access_ro]

        self._mock_powerstore_client.get_nfs_export_id.return_value = (
            self.NFS_EXPORT_ID
        )

        access_updates = self.storage_connection.update_access(
            self.mock_context,
            share,
            access_rules,
            add_rules=None,
            delete_rules=None,
            share_server=None,
        )

        self._mock_powerstore_client.set_export_access.assert_called_once_with(
            self.NFS_EXPORT_ID, set(), set()
        )

        self.assertIsNotNone(access_updates)

    def test_update_access_add_cifs_invalid_acess_type(self):
        share = {
            "name": self.SHARE_NAME,
            "share_proto": "CIFS",
            "display_name": "foo_display_name",
        }

        cifs_access_rw = {
            "access_type": "invalid_type",
            "access_to": self.RW_USERS,
            "access_level": const.ACCESS_LEVEL_RW,
            "access_id": "09960614-8574-4e03-89cf-7cf267b0bd08",
        }
        cifs_access_ro = {
            "access_type": "invalid_type",
            "access_to": self.RO_USERS,
            "access_level": const.ACCESS_LEVEL_RO,
            "access_id": "09960614-8574-4e03-89cf-7cf267b0bd09",
        }
        access_rules = [cifs_access_rw, cifs_access_ro]

        self._mock_powerstore_client.get_smb_share_id.return_value = (
            self.SMB_SHARE_ID
        )

        access_updates = self.storage_connection.update_access(
            self.mock_context,
            share,
            access_rules,
            add_rules=None,
            delete_rules=None,
            share_server=None,
        )

        self._mock_powerstore_client.set_acl.assert_called_once_with(
            self.SMB_SHARE_ID, set(), set()
        )

        self.assertIsNotNone(access_updates)

    def test_update_access_add_nfs_backend_failure(self):
        share = {
            "name": self.SHARE_NAME,
            "share_proto": "NFS",
            "display_name": "foo_display_name",
        }

        self._mock_powerstore_client.get_nfs_export_id.return_value = (
            self.NFS_EXPORT_ID
        )
        self._mock_powerstore_client.set_export_access.return_value = False

        self.assertFalse(self._mock_powerstore_client.get_nfs_export_id.called)
        self.assertFalse(self._mock_powerstore_client.set_export_access.called)

        nfs_access_rw = {
            "access_type": "ip",
            "access_to": self.RW_HOSTS,
            "access_level": const.ACCESS_LEVEL_RW,
            "access_id": "09960614-8574-4e03-89cf-7cf267b0bd08",
        }
        nfs_access_ro = {
            "access_type": "ip",
            "access_to": self.RO_HOSTS,
            "access_level": const.ACCESS_LEVEL_RO,
            "access_id": "09960614-8574-4e03-89cf-7cf267b0bd08",
        }
        access_rules = [nfs_access_rw, nfs_access_ro]

        self.assertRaises(
            exception.ShareBackendException,
            self.storage_connection.update_access,
            self.mock_context,
            share,
            access_rules,
            add_rules=None,
            delete_rules=None,
            share_server=None,
        )

    def test_update_access_add_cifs_backend_failure(self):
        share = {
            "name": self.SHARE_NAME,
            "share_proto": "CIFS",
            "display_name": "foo_display_name",
        }

        self._mock_powerstore_client.set_acl.return_value = False

        cifs_access_rw = {
            "access_type": "user",
            "access_to": self.RW_USERS,
            "access_level": const.ACCESS_LEVEL_RW,
            "access_id": "09960614-8574-4e03-89cf-7cf267b0bd08",
        }
        cifs_access_ro = {
            "access_type": "user",
            "access_to": self.RO_USERS,
            "access_level": const.ACCESS_LEVEL_RO,
            "access_id": "09960614-8574-4e03-89cf-7cf267b0bd08",
        }
        access_rules = [cifs_access_rw, cifs_access_ro]

        self.assertRaises(
            exception.ShareBackendException,
            self.storage_connection.update_access,
            self.mock_context,
            share,
            access_rules,
            add_rules=None,
            delete_rules=None,
            share_server=None,
        )

    def test_allow_access(self):
        self.assertRaises(
            NotImplementedError,
            self.storage_connection.allow_access,
            self.mock_context,
            share=None,
            access=None,
            share_server=None,
        )

    def test_deny_access(self):
        self.assertRaises(
            NotImplementedError,
            self.storage_connection.deny_access,
            self.mock_context,
            share=None,
            access=None,
            share_server=None,
        )

    def test_update_share_stats(self):
        data = dict(
            share_backend_name='powerstore',
            vendor_name='Dell EMC',
            storage_protocol='NFS_CIFS',
            snapshot_support=True,
            create_share_from_snapshot_support=True)
        self._mock_powerstore_client.get_cluster_id.return_value = "0"
        self._mock_powerstore_client. \
            retreive_cluster_capacity_metrics.return_value = \
            47345047046144, 366003363027
        self.storage_connection.update_share_stats(data)
        self.assertEqual(data['storage_protocol'], 'NFS_CIFS')
        self.assertEqual(data['driver_version'], connection.VERSION)
        self.assertEqual(data['total_capacity_gb'], 44093)
        self.assertEqual(data['free_capacity_gb'], 43752)

    def test_update_share_stats_failure(self):
        data = dict(
            share_backend_name='powerstore',
            vendor_name='Dell EMC',
            storage_protocol='NFS_CIFS',
            snapshot_support=True,
            create_share_from_snapshot_support=True)
        self._mock_powerstore_client. \
            retreive_cluster_capacity_metrics.return_value = \
            None, None
        self.storage_connection.update_share_stats(data)
        self.assertIsNone(data.get('total_capacity_gb'))
        self.assertIsNone(data.get('free_capacity_gb'))

    def test_create_snapshot(self):
        self._mock_powerstore_client.get_filesystem_id.return_value = (
            self.FILESYSTEM_ID
        )
        self._mock_powerstore_client. \
            create_snapshot.return_value = self.SNAPSHOT_ID

        snapshot = {
            "name": self.SNAPSHOT_NAME,
            "share_name": self.SHARE_NAME
        }
        self.storage_connection.create_snapshot(
            self.mock_context, snapshot, None
        )

        self._mock_powerstore_client.get_filesystem_id. \
            assert_called_with(
                self.SHARE_NAME
            )
        self._mock_powerstore_client.create_snapshot.assert_called_with(
            self.FILESYSTEM_ID, self.SNAPSHOT_NAME
        )

    def test_create_snapshot_invalid_filesystem_id(self):
        self._mock_powerstore_client.get_filesystem_id.return_value = (
            None
        )

        snapshot = {
            "name": self.SNAPSHOT_NAME,
            "share_name": self.SHARE_NAME
        }
        self.assertRaises(
            exception.ShareBackendException,
            self.storage_connection.create_snapshot,
            self.mock_context,
            snapshot,
            None
        )

    def test_create_snapshot_backend_failure(self):
        self._mock_powerstore_client.get_filesystem_id.return_value = (
            self.FILESYSTEM_ID
        )
        self._mock_powerstore_client. \
            create_snapshot.return_value = None

        snapshot = {
            "name": self.SNAPSHOT_NAME,
            "share_name": self.SHARE_NAME,
            "share": {
                "share_proto": "NFS"
            }
        }
        self.assertRaises(
            exception.ShareBackendException,
            self.storage_connection.create_snapshot,
            self.mock_context,
            snapshot,
            None
        )

    def test_delete_snapshot(self):
        self._mock_powerstore_client.get_filesystem_id.return_value = (
            self.SNAPSHOT_ID
        )
        self._mock_powerstore_client.delete_filesystem.return_value = True

        snapshot = {
            "name": self.SNAPSHOT_NAME,
            "share_name": self.SHARE_NAME
        }
        self.storage_connection.delete_snapshot(
            self.mock_context, snapshot, None
        )

        self._mock_powerstore_client.get_filesystem_id. \
            assert_called_with(
                self.SNAPSHOT_NAME
            )
        self._mock_powerstore_client.delete_filesystem.assert_called_with(
            self.SNAPSHOT_ID
        )

    def test_delete_snapshot_backend_failure(self):
        self._mock_powerstore_client.get_filesystem_id.return_value = (
            self.SNAPSHOT_ID
        )
        self._mock_powerstore_client.delete_filesystem.return_value = False

        snapshot = {
            "name": self.SNAPSHOT_NAME,
            "share_name": self.SHARE_NAME
        }
        self.assertRaises(
            exception.ShareBackendException,
            self.storage_connection.delete_snapshot,
            self.mock_context,
            snapshot,
            None
        )

    def test_revert_to_snapshot(self):
        self._mock_powerstore_client.get_filesystem_id.return_value = (
            self.SNAPSHOT_ID
        )
        self._mock_powerstore_client.restore_snapshot.return_value = True

        snapshot = {
            "name": self.SNAPSHOT_NAME,
            "share_name": self.SHARE_NAME
        }
        self.storage_connection.revert_to_snapshot(
            self.mock_context, snapshot, None, None, None
        )

        self._mock_powerstore_client.get_filesystem_id. \
            assert_called_with(
                self.SNAPSHOT_NAME
            )
        self._mock_powerstore_client.restore_snapshot.assert_called_with(
            self.SNAPSHOT_ID
        )

    def test_revert_to_snapshot_backend_failure(self):
        self._mock_powerstore_client.get_filesystem_id.return_value = (
            self.SNAPSHOT_ID
        )
        self._mock_powerstore_client.restore_snapshot.return_value = False

        snapshot = {
            "name": self.SNAPSHOT_NAME,
            "share_name": self.SHARE_NAME
        }
        self.assertRaises(
            exception.ShareBackendException,
            self.storage_connection.revert_to_snapshot,
            self.mock_context,
            snapshot,
            None, None, None
        )

    def test_create_share_from_snapshot_nfs(self):
        self._mock_powerstore_client.get_nas_server_id.return_value = (
            self.NAS_SERVER_ID
        )
        self._mock_powerstore_client.get_filesystem_id.return_value = (
            self.SNAPSHOT_ID
        )
        self._mock_powerstore_client.clone_snapshot.return_value = (
            self.CLONE_ID
        )
        self._mock_powerstore_client.create_nfs_export.return_value = (
            self.NFS_EXPORT_ID
        )
        self._mock_powerstore_client.get_nas_server_interfaces.return_value = (
            [{"ip": self.NAS_SERVER_IP, "preferred": True}]
        )
        self._mock_powerstore_client.resize_filesystem.return_value = (
            True, None
        )
        share = {"name": self.SHARE_NAME, "share_proto": "NFS",
                 "size": self.SHARE_NEW_SIZE_GB}
        snapshot = {"name": self.SNAPSHOT_NAME, "size": self.SHARE_SIZE_GB}
        locations = self.storage_connection.create_share_from_snapshot(
            self.mock_context,
            share,
            snapshot
        )
        expected_locations = [
            {"path": "%s:/%s" % (
                self.NAS_SERVER_IP,
                self.SHARE_NAME,
                ),
             "metadata": {
                 "preferred": True}}]
        self.assertEqual(expected_locations, locations)
        self._mock_powerstore_client.get_nas_server_id.assert_called_with(
            self._mock_config.safe_get("dell_nas_server")
        )
        self._mock_powerstore_client.clone_snapshot.assert_called_with(
            self.SNAPSHOT_ID,
            self.SHARE_NAME
        )
        self._mock_powerstore_client.create_nfs_export.assert_called_with(
            self.CLONE_ID,
            self.SHARE_NAME
        )
        self._mock_powerstore_client.get_nas_server_interfaces. \
            assert_called_with(
                self.NAS_SERVER_ID
            )

    def test_create_share_from_snapshot_cifs(self):
        self._mock_powerstore_client.get_nas_server_id.return_value = (
            self.NAS_SERVER_ID
        )
        self._mock_powerstore_client.get_filesystem_id.return_value = (
            self.SNAPSHOT_ID
        )
        self._mock_powerstore_client.clone_snapshot.return_value = (
            self.CLONE_ID
        )
        self._mock_powerstore_client.create_smb_share.return_value = (
            self.NFS_EXPORT_ID
        )
        self._mock_powerstore_client.get_nas_server_interfaces.return_value = (
            [{"ip": self.NAS_SERVER_IP, "preferred": True}]
        )
        self._mock_powerstore_client.resize_filesystem.return_value = (
            True, None
        )

        share = {"name": self.SHARE_NAME, "share_proto": "CIFS",
                 "size": self.SHARE_NEW_SIZE_GB}
        snapshot = {"name": self.SNAPSHOT_NAME, "size": self.SHARE_SIZE_GB}
        locations = self.storage_connection.create_share_from_snapshot(
            self.mock_context,
            share,
            snapshot
        )
        expected_locations = [
            {"path": "\\\\%s\\%s" % (
                self.NAS_SERVER_IP,
                self.SHARE_NAME),
             "metadata": {
                 "preferred": True}}]
        self.assertEqual(expected_locations, locations)
        self._mock_powerstore_client.get_nas_server_id.assert_called_with(
            self._mock_config.safe_get("dell_nas_server")
        )
        self._mock_powerstore_client.clone_snapshot.assert_called_with(
            self.SNAPSHOT_ID,
            self.SHARE_NAME
        )
        self._mock_powerstore_client.create_smb_share.assert_called_with(
            self.CLONE_ID,
            self.SHARE_NAME
        )
        self._mock_powerstore_client.get_nas_server_interfaces. \
            assert_called_with(
                self.NAS_SERVER_ID
            )

    def test_create_share_from_snapshot_clone_failure(self):
        share = {"name": self.SHARE_NAME, "share_proto": "NFS"}
        snapshot = {"name": self.SNAPSHOT_NAME}
        self._mock_powerstore_client.clone_snapshot.return_value = None

        self.assertRaises(
            exception.ShareBackendException,
            self.storage_connection.create_share_from_snapshot,
            self.mock_context,
            share,
            snapshot
        )

    def test_create_share_from_snapshot_export_failure(self):
        share = {"name": self.SHARE_NAME, "share_proto": "NFS"}
        snapshot = {"name": self.SNAPSHOT_NAME}
        self._mock_powerstore_client.create_nfs_export.return_value = None

        self.assertRaises(
            exception.ShareBackendException,
            self.storage_connection.create_share_from_snapshot,
            self.mock_context,
            share,
            snapshot
        )

    def test_create_share_from_snapshot_share_failure(self):
        share = {"name": self.SHARE_NAME, "share_proto": "CIFS"}
        snapshot = {"name": self.SNAPSHOT_NAME}
        self._mock_powerstore_client.create_smb_share.return_value = None

        self.assertRaises(
            exception.ShareBackendException,
            self.storage_connection.create_share_from_snapshot,
            self.mock_context,
            share,
            snapshot
        )

    def test_get_default_filter_function(self):
        filter = self.storage_connection.get_default_filter_function()
        self.assertEqual(filter, "share.size >= 3")



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import versionutils
from oslo_policy import policy

from manila.policies import base


BASE_POLICY_NAME = 'share_snapshot:%s'

DEPRECATED_REASON = """
The share snapshot API now supports scope and default roles.
"""

deprecated_snapshot_get = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'get_snapshot',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_snapshot_get_all = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'get_all_snapshots',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_snapshot_force_delete = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'force_delete',
    check_str=base.RULE_ADMIN_API,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_snapshot_manage = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'manage_snapshot',
    check_str=base.RULE_ADMIN_API,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_snapshot_unmanage = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'unmanage_snapshot',
    check_str=base.RULE_ADMIN_API,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_snapshot_reset_status = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'reset_status',
    check_str=base.RULE_ADMIN_API,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_snapshot_access_list = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'access_list',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_snapshot_allow_access = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'allow_access',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_snapshot_deny_access = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'deny_access',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_update_snapshot_metadata = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'update_metadata',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since='ZED'
)
deprecated_delete_snapshot_metadata = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'delete_metadata',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since='ZED'
)
deprecated_get_snapshot_metadata = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'get_metadata',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since='ZED'
)
deprecated_list_snapshots_in_deferred_deletion_states = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'list_snapshots_in_deferred_deletion_states',
    check_str=base.RULE_ADMIN_API,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since='2024.1/Caracal'
)

share_snapshot_policies = [
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'get_snapshot',
        check_str=base.ADMIN_OR_PROJECT_READER,
        scope_types=['project'],
        description="Get share snapshot.",
        operations=[
            {
                'method': 'GET',
                'path': '/snapshots/{snapshot_id}'
            }
        ],
        deprecated_rule=deprecated_snapshot_get
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'get_all_snapshots',
        check_str=base.ADMIN_OR_PROJECT_READER,
        scope_types=['project'],
        description="Get all share snapshots.",
        operations=[
            {
                'method': 'GET',
                'path': '/snapshots?{query}'
            },
            {
                'method': 'GET',
                'path': '/snapshots/detail?{query}'
            }
        ],
        deprecated_rule=deprecated_snapshot_get_all
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'force_delete',
        check_str=base.ADMIN,
        scope_types=['project'],
        description="Force Delete a share snapshot.",
        operations=[
            {
                'method': 'DELETE',
                'path': '/snapshots/{snapshot_id}'
            }
        ],
        deprecated_rule=deprecated_snapshot_force_delete
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'manage_snapshot',
        check_str=base.ADMIN,
        scope_types=['project'],
        description="Manage share snapshot.",
        operations=[
            {
                'method': 'POST',
                'path': '/snapshots/manage'
            }
        ],
        deprecated_rule=deprecated_snapshot_manage
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'unmanage_snapshot',
        check_str=base.ADMIN,
        scope_types=['project'],
        description="Unmanage share snapshot.",
        operations=[
            {
                'method': 'POST',
                'path': '/snapshots/{snapshot_id}/action'
            }
        ],
        deprecated_rule=deprecated_snapshot_unmanage
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'reset_status',
        check_str=base.ADMIN,
        scope_types=['project'],
        description="Reset status.",
        operations=[
            {
                'method': 'POST',
                'path': '/snapshots/{snapshot_id}/action',
            }
        ],
        deprecated_rule=deprecated_snapshot_reset_status
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'access_list',
        check_str=base.ADMIN_OR_PROJECT_READER,
        scope_types=['project'],
        description="List access rules of a share snapshot.",
        operations=[
            {
                'method': 'GET',
                'path': '/snapshots/{snapshot_id}/access-list'
            }
        ],
        deprecated_rule=deprecated_snapshot_access_list
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'allow_access',
        check_str=base.ADMIN_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description="Allow access to a share snapshot.",
        operations=[
            {
                'method': 'POST',
                'path': '/snapshots/{snapshot_id}/action'
            }
        ],
        deprecated_rule=deprecated_snapshot_allow_access
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'deny_access',
        check_str=base.ADMIN_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description="Deny access to a share snapshot.",
        operations=[
            {
                'method': 'POST',
                'path': '/snapshots/{snapshot_id}/action'
            }
        ],
        deprecated_rule=deprecated_snapshot_deny_access
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'update_metadata',
        check_str=base.ADMIN_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description="Update snapshot metadata.",
        operations=[
            {
                'method': 'PUT',
                'path': '/snapshots/{snapshot_id}/metadata',
            },
            {
                'method': 'POST',
                'path': '/snapshots/{snapshot_id}/metadata/{key}',
            },
            {
                'method': 'POST',
                'path': '/snapshots/{snapshot_id}/metadata',
            },
        ],
        deprecated_rule=deprecated_update_snapshot_metadata
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'delete_metadata',
        check_str=base.ADMIN_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description="Delete snapshot metadata.",
        operations=[
            {
                'method': 'DELETE',
                'path': '/snapshots/{snapshot_id}/metadata/{key}',
            }
        ],
        deprecated_rule=deprecated_delete_snapshot_metadata
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'get_metadata',
        check_str=base.ADMIN_OR_PROJECT_READER,
        scope_types=['project'],
        description="Get snapshot metadata.",
        operations=[
            {
                'method': 'GET',
                'path': '/snapshots/{snapshot_id}/metadata',
            },
            {
                'method': 'GET',
                'path': '/snapshots/{snapshot_id}/metadata/{key}',
            }
        ],
        deprecated_rule=deprecated_get_snapshot_metadata
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'list_snapshots_in_deferred_deletion_states',
        check_str=base.ADMIN,
        scope_types=['project'],
        description="List (or get) snapshots whose deletion has been deferred",
        operations=[
            {
                'method': 'GET',
                'path': '/v2/snapshots',
            },
            {
                'method': 'GET',
                'path': '/snapshots/{snapshot_id}'
            }
        ],
        deprecated_rule=deprecated_list_snapshots_in_deferred_deletion_states
    ),
]


def list_rules():
    return share_snapshot_policies



# Copyright 2023 NetApp, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Tests For NetApp Active IQ Weigher.
"""

from unittest import mock

import ddt
from oslo_config import cfg
from oslo_serialization import jsonutils
import requests

from manila import context
from manila import exception
from manila.scheduler.weighers import base_host
from manila.scheduler.weighers import netapp_aiq
from manila.share import utils
from manila import test
from manila.tests.scheduler import fakes
from manila.tests import utils as test_utils

CONF = cfg.CONF


@ddt.ddt
class NetAppAIQWeigherTestCase(test.TestCase):
    def setUp(self):
        super(NetAppAIQWeigherTestCase, self).setUp()
        self.weight_handler = base_host.HostWeightHandler(
            'manila.scheduler.weighers')

        netapp_aiq.LOG.debug = mock.Mock()
        netapp_aiq.LOG.error = mock.Mock()

        self.mock_session = mock.Mock()
        self.mock_session.get = mock.Mock()
        self.mock_session.post = mock.Mock()
        self.mock_session.delete = mock.Mock()
        self.mock_session.patch = mock.Mock()
        self.mock_session.put = mock.Mock()

        data = {
            'netapp_active_iq': {
                'aiq_hostname': "10.10.10.10",
                'aiq_transport_type': 'https',
                'aiq_ssl_verify': True,
                'aiq_ssl_cert_path': 'fake_cert',
                'aiq_username': 'fake_user',
                'aiq_password': 'fake_password',
                'aiq_eval_method': 1,
                'aiq_priority_order': 'ops'
            }
        }
        self.netapp_aiq_weigher = None
        with test_utils.create_temp_config_with_opts(data):
            self.netapp_aiq_weigher = netapp_aiq.NetAppAIQWeigher()

    def test__weigh_object(self):
        self.assertRaises(NotImplementedError,
                          self.netapp_aiq_weigher._weigh_object,
                          "fake", "fake")

    @ddt.data(
        {'resource_keys': ["fake_resource_key"], 'performance_level': None},
        {'resource_keys': ["fake_resource_key"],
         'performance_level': "fake_psl"},
        {'resource_keys': [], 'performance_level': 'fake_psl'})
    @ddt.unpack
    def test__weigh_active_iq(self, resource_keys, performance_level):
        weight_properties = {
            'size': 1,
            'share_type': {
                'extra_specs': {
                    "netapp:performance_service_level_name": "fake_name",
                }
            }
        }
        mock_get_psl_id = self.mock_object(
            self.netapp_aiq_weigher, '_get_performance_level_id',
            mock.Mock(return_value=performance_level))
        mock_get_resource_keys = self.mock_object(
            self.netapp_aiq_weigher, '_get_resource_keys',
            mock.Mock(return_value=resource_keys))
        mock_balance_aggregates = self.mock_object(
            self.netapp_aiq_weigher, '_balance_aggregates',
            mock.Mock(return_value=["1.0", "1.0"]))

        res = self.netapp_aiq_weigher._weigh_active_iq(
            fakes.FAKE_ACTIVE_IQ_WEIGHER_LIST, weight_properties)

        mock_get_psl_id.assert_called_once_with("fake_name")
        if not resource_keys or not performance_level:
            self.assertEqual([], res)
        else:
            self.assertEqual(["1.0", "1.0"], res)
        if performance_level:
            mock_get_resource_keys.assert_called_once_with(
                fakes.FAKE_ACTIVE_IQ_WEIGHER_LIST)
        else:
            mock_get_resource_keys.assert_not_called()
        if not resource_keys or not performance_level:
            mock_balance_aggregates.assert_not_called()
        else:
            mock_balance_aggregates.assert_called_once_with(
                resource_keys, 1, performance_level)

    @ddt.data(True, False)
    def test__get_url(self, ipv6):
        if ipv6:
            self.netapp_aiq_weigher.host = "2001:db8::"
        else:
            self.netapp_aiq_weigher.host = "1.1.1.1"
        self.netapp_aiq_weigher.port = "fake_port"
        self.netapp_aiq_weigher.protocol = "fake_protocol"

        res = self.netapp_aiq_weigher._get_url()

        if ipv6:
            self.assertEqual('fake_protocol://[2001:db8::]:fake_port/api/',
                             res)
        else:
            self.assertEqual('fake_protocol://1.1.1.1:fake_port/api/',
                             res)

    @ddt.data('get', 'post', 'delete', 'patch', 'put')
    def test__get_request_method(self, method):
        res = self.netapp_aiq_weigher._get_request_method(
            method, self.mock_session)

        if method == 'get':
            self.assertEqual(self.mock_session.get, res)
        elif method == 'post':
            self.assertEqual(self.mock_session.post, res)
        elif method == 'delete':
            self.assertEqual(self.mock_session.delete, res)
        elif method == 'put':
            self.assertEqual(self.mock_session.put, res)
        elif method == 'patch':
            self.assertEqual(self.mock_session.patch, res)

    def test__get_session_method(self):
        mock_session_builder = self.mock_object(
            requests, 'Session', mock.Mock(return_value=self.mock_session))
        mock__get_request_method = self.mock_object(
            self.netapp_aiq_weigher, '_get_request_method',
            mock.Mock(return_value=self.mock_session.post))

        res = self.netapp_aiq_weigher._get_session_method('post')

        self.assertEqual(self.mock_session.post, res)
        mock_session_builder.assert_called_once_with()
        mock__get_request_method.assert_called_once_with(
            'post', self.mock_session)

    def test__call_active_iq(self):
        response = mock.Mock()
        response.content = "fake_response"
        response.status_code = "fake_code"
        mock_post = mock.Mock(return_value=response)
        mock__get_session_method = self.mock_object(
            self.netapp_aiq_weigher, '_get_session_method',
            mock.Mock(return_value=mock_post))
        fake_url = "fake_url"
        fake_path = "/fake_path"
        mock__get_url = self.mock_object(
            self.netapp_aiq_weigher, '_get_url',
            mock.Mock(return_value=fake_url))

        self.netapp_aiq_weigher._call_active_iq(fake_path, "post",
                                                body="fake_body")

        mock_post.assert_called_once_with(fake_url + fake_path,
                                          json="fake_body")
        self.assertTrue(netapp_aiq.LOG.debug.called)
        mock__get_session_method.assert_called_once_with("post")
        mock__get_url.assert_called_once_with()

    @ddt.data({}, jsonutils.dumps(
        fakes.FAKE_ACTIVE_IQ_WEIGHER_AGGREGATES_RESPONSE))
    def test__get_resource_keys(self, api_res):
        mock__call_active_iq = self.mock_object(
            self.netapp_aiq_weigher, '_call_active_iq',
            mock.Mock(return_value=(200, api_res)))

        res = self.netapp_aiq_weigher._get_resource_keys(
            fakes.FAKE_ACTIVE_IQ_WEIGHER_LIST)

        if api_res:
            self.assertEqual(['fake_key_1', 'fake_key_2', 'fake_key_3'], res)
        else:
            self.assertEqual([0, 0, 0], res)
        mock__call_active_iq.assert_called_once_with(
            'datacenter/storage/aggregates', 'get')

    @ddt.data(mock.Mock(side_effect=exception.NotFound),
              mock.Mock(return_value=(400, "fake_res")))
    def test__get_resource_keys_error(self, mock_cal):
        self.mock_object(
            self.netapp_aiq_weigher, '_call_active_iq', mock_cal)

        res = self.netapp_aiq_weigher._get_resource_keys(
            fakes.FAKE_ACTIVE_IQ_WEIGHER_LIST)

        self.assertEqual([], res)
        self.assertTrue(netapp_aiq.LOG.error.called)

    @ddt.data([], jsonutils.dumps(
        fakes.FAKE_ACTIVE_IQ_WEIGHER_BALANCE_RESPONSE))
    def test__balance_aggregates(self, api_res):
        mock__call_active_iq = self.mock_object(
            self.netapp_aiq_weigher, '_call_active_iq',
            mock.Mock(return_value=(200, api_res)))

        res = self.netapp_aiq_weigher._balance_aggregates(
            ['fake_key_1', 'fake_key_2', 0, 'fake_key_3'], 10, 'fake_uuid')

        if not api_res:
            self.assertEqual([0.0, 0.0, 0.0, 0.0], res)
        else:
            self.assertEqual([10.0, 20.0, 0.0, 0.0], res)
        fake_body = {
            "capacity": '10GB',
            "eval_method": 1,
            "opt_method": 0,
            "priority_order": ['ops'],
            "separate_flag": False,
            "resource_keys": ['fake_key_1', 'fake_key_2', 'fake_key_3'],
            "ssl_key": 'fake_uuid'
        }
        mock__call_active_iq.assert_called_once_with(
            'storage-provider/data-placement/balance', 'post', body=fake_body)

    @ddt.data(mock.Mock(side_effect=exception.NotFound),
              mock.Mock(return_value=(400, "fake_res")))
    def test__balance_aggregates_error(self, mock_cal):
        self.mock_object(
            self.netapp_aiq_weigher, '_call_active_iq', mock_cal)

        res = self.netapp_aiq_weigher._balance_aggregates(
            ['fake_key_1', 'fake_key_2', 0, 'fake_key_3'], 10, 'fake_uuid')

        self.assertEqual([], res)
        self.assertTrue(netapp_aiq.LOG.error.called)

    @mock.patch('manila.db.api.IMPL.service_get_all_by_topic')
    def _get_all_hosts(self, _mock_service_get_all_by_topic, disabled=False):
        ctxt = context.get_admin_context()
        fakes.mock_host_manager_db_calls(_mock_service_get_all_by_topic,
                                         disabled=disabled)
        host_states = self.host_manager.get_all_host_states_share(ctxt)
        _mock_service_get_all_by_topic.assert_called_once_with(
            ctxt, CONF.share_topic, consider_disabled=False)
        return host_states

    def test_weigh_objects_netapp_only(self):
        self.host_manager = fakes.FakeHostManagerNetAppOnly()
        hosts = self._get_all_hosts()  # pylint: disable=no-value-for-parameter
        weight_properties = "fake_properties"
        mock_weigh_active_iq = self.mock_object(
            netapp_aiq.NetAppAIQWeigher, '_weigh_active_iq',
            # third host wins
            mock.Mock(return_value=[0.0, 0.0, 10.0, 0.0, 0.0, 0.0]))

        weighed_host = self.weight_handler.get_weighed_objects(
            [netapp_aiq.NetAppAIQWeigher],
            hosts,
            weight_properties)[0]

        mock_weigh_active_iq.assert_called()
        self.assertEqual(1.0, weighed_host.weight)
        self.assertEqual(
            'host3', utils.extract_host(weighed_host.obj.host))

    def test_weigh_objects_non_netapp_backends(self):
        self.host_manager = fakes.FakeHostManager()
        hosts = self._get_all_hosts()  # pylint: disable=no-value-for-parameter
        weight_properties = "fake_properties"
        mock_weigh_active_iq = self.mock_object(
            netapp_aiq.NetAppAIQWeigher, '_weigh_active_iq')

        weighed_host = self.weight_handler.get_weighed_objects(
            [netapp_aiq.NetAppAIQWeigher],
            hosts,
            weight_properties)[0]

        mock_weigh_active_iq.assert_not_called()
        self.assertEqual(0.0, weighed_host.weight)
        self.assertEqual(
            'host1', utils.extract_host(weighed_host.obj.host))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""add-ensuring-field-to-services

Revision ID: cdefa6287df8
Revises: 2f27d904214c
Create Date: 2024-07-15 14:29:16.733696

"""

# revision identifiers, used by Alembic.
revision = 'cdefa6287df8'
down_revision = '2f27d904214c'

from alembic import op
from oslo_log import log
import sqlalchemy as sa


LOG = log.getLogger(__name__)


def upgrade():
    try:
        op.add_column('services', sa.Column(
            'ensuring', sa.Boolean,
            nullable=False, server_default=sa.sql.false()))
    except Exception:
        LOG.error("Column services.ensuring not created!")
        raise


def downgrade():
    try:
        op.drop_column('services', 'ensuring')
    except Exception:
        LOG.error("Column shares.ensuring not dropped!")
        raise
